#!/usr/local/bin/perl5.6.0 -w

# Constraint variables declaration before using them
use strict;
##########################################

# Bio modules
# PSU
use Bio::PSU::IO::Blast::Result;
use Bio::PSU::IO::Blast::Search;
use Bio::PSU::SeqFactory;
use Bio::PSU::Range;
use Bio::PSU::Feature;

use Cwd;
##########################################

# Benchmark Module
use Benchmark;
##################################################################

use Data::Dumper;
##################################################################

=head1 NAME

annotations_update.pl - help the annotations of unfinished sequences 
  - read embl/tab files 
  - mapping the annotations from a coordinates system to another one
  - reporting discrepancies

=head1 SYNOPSIS

perl ~/bin/annotations_update.pl -help -oldseq [old sequence fasta file] -newseq [new sequence fasta file] -tab [tab file] -speedup [true|false] -filtering [blast score]
 
  -help    a help message
  -tab     the tab file
  -seqold  the previous sequence fasta file 
  -seqnew  the new sequence fasta file
  -speedup to speedup the blast searches [true|false]
  -filtering to filter features which must not be mapped (e.g. can't be in the new sequence)

=head1 DESCRIPTION

  The output is a new tab file with the suffix ".new"

  @author Arnaud Kerhornou axk@sanger.ac.uk
  
  PSU BioModules : Keith James kdj@sanger.ac.uk

=cut

my $t1 = Benchmark->new ();

my $debug = 0;

my $_help            = "-help";
my $_tabfile         = "-tab";
my $_oldseqfile      = "-oldseq";
my $_newseqfile      = "-newseq";
my $_speedup         = "-speedup";
my $_filtering       = "-filtering";

my $tabfile          = "-1";
my $oldseqfile       = "-1";
my $newseqfile       = "-1";
my $speedup          = "true";
my $filtering        = 0;

# Check if in the command line there is an input file name (-i option)
# the other options are not compulsory (set of options per default)

my $i = 0;
foreach my $element (@ARGV) {
  if ($element eq $_help) {
    display_help ();
    exit 1;
  }
  elsif ($element eq $_tabfile) {
    $tabfile = $ARGV[$i+1];
  }
  elsif ($element eq $_oldseqfile) {
    $oldseqfile = $ARGV[$i+1];
  }
  elsif ($element eq $_newseqfile) {
    $newseqfile = $ARGV[$i+1];
  }
  elsif ($element eq $_speedup) {
    $speedup = $ARGV[$i+1];
  }
  elsif ($element eq $_filtering) {
    $filtering = $ARGV[$i+1];
  }
  # if at least one of the options is not recognized, display the help message and exit
  elsif ($element =~ /^-/) {
    print STDERR "option not recognized: $element\n";
    display_help ();
    exit 1;
  }
  $i ++;
}
if ($oldseqfile eq "-1" or $newseqfile eq "-1") {
  display_help ();
  exit 1;
}

my $taboutput      = "mappedAnnotations.embl";
my $failedfeatfile = "failures.tab";

my $curdir = cwd;
$curdir =~ s/\/tmp_mnt//g;

# my $oldseqfile_path = $curdir . "/" . $oldseqfile;
# my $newseqfile_path = $curdir . "/" . $newseqfile;

print STDERR "loading the sequence and feature objects from files ...\n";

###########################
# Get the old sequence object
# the annotations are within this object or within a tab file
##

my $old_s1 = Bio::PSU::SeqFactory->make(-file   => "<$oldseqfile");
my $oldseq_obj = $old_s1->next_seq;

# check if there are annotations to map !!

if ((! $oldseq_obj->has_features) and ($tabfile eq "-1")) {
  print STDERR "No annotations to map !!\n";
  display_help ();
  exit 1;
}

###########################
# Get the new sequence object
##

my $new_s1     = Bio::PSU::SeqFactory->make(-file   => "<$newseqfile");
my $newseq_obj = $new_s1->next_seq;

my @new_features    = ();
my @failed_features = ();

if ($newseq_obj->has_features) {
  push (@new_features, $newseq_obj->features);
}

# the tab object

my $tabfile_path = $curdir . '/' . $tabfile;
my $tab_seq_obj;
if (! ($tabfile eq "-1")) {
  my $tab_obj = Bio::PSU::SeqFactory->make(-file   => "<$tabfile");
  $tab_seq_obj = $tab_obj->next_seq;  
}
else {
  $tab_seq_obj = $oldseq_obj;
}

##########################################
# to check with an EMBL file ...
##########################################

my $add_sequence = 1;
if (($tabfile ne "-1") and (not defined ($tab_seq_obj->str))) {
  $add_sequence = 0;
}

#####################################################
# Mapping ...
#####################################################

my @args = ("mkdir blast_search");
system (@args);

if ((my $nb_features = $tab_seq_obj->has_features) > 0) {
  print STDERR "$nb_features features are being mapped ...\n\n";
  my $i = 1;
  my $percent_memorized = -1;

  foreach my $feature ($tab_seq_obj->features) {

    print STDERR "mapping new feature ...\n";
    
    my $new_feature = mapping ($feature, $oldseq_obj, $newseq_obj, $curdir, $speedup, $filtering);
    
    # if not failed
    # cloning the current feature to create a new one, or working on the current one, just modyfying the set of ranges
    # otherwise copy it into failures.tab
    
    if (defined ($new_feature)) {
      push (@new_features, $new_feature);
    }
    else {
      push (@failed_features, $feature);
    }
    
    my $percent = int ($i*100/$nb_features);
    if ((($percent%5) == 0) && ($percent != $percent_memorized)) {
      print STDERR "\n$percent% done.\n\n";
      $percent_memorized = $percent;
    }
    $i++;
  }
}

#####################################################
# Mapping done
# copy the new set of features into a new tab file
# copy the failed features into the failure tab file
#####################################################

# if embl format file, there is a sequence
# else this is a tab file, no sequence

my $new_tab_seq_obj = undef;

if ($add_sequence) {
  $new_tab_seq_obj = Bio::PSU::Seq->new (-id   => $tab_seq_obj->id,
					 -desc => $tab_seq_obj->desc,
					 -type => $tab_seq_obj->type,
					 -str  => $newseq_obj->str
					);
}
else {
  $new_tab_seq_obj = Bio::PSU::Seq->new (-id   => $tab_seq_obj->id,
					 -desc => $tab_seq_obj->desc,
					 -type => $tab_seq_obj->type
					);
}
$new_tab_seq_obj->features (@new_features);

# copy this seq obj into a new file

my $taboutput_path = $curdir . '/' . $taboutput;
my $new_tab_obj = Bio::PSU::SeqFactory->make(-file   => ">$taboutput",
					     -format => 'embl');
$new_tab_obj->write_seq($new_tab_seq_obj);

# failed features

my $failed_tab_seq_obj= Bio::PSU::Seq->new (-id   => $tab_seq_obj->id,
					    -desc => $tab_seq_obj->desc,
					    -type => $tab_seq_obj->type);
$failed_tab_seq_obj->features (@failed_features);

# copy this seq obj into a new file

my $failedfeatfile_path = $curdir . '/' . $failedfeatfile;
my $failed_tab_obj = Bio::PSU::SeqFactory->make(-file   => ">$failedfeatfile",
						-format => 'embl');
$failed_tab_obj->write_seq($failed_tab_seq_obj);

my $t2 = Benchmark->new ();
print  STDERR "\nTotal : ", timestr (timediff ($t2, $t1)), "\n";

###################################################################
# The end
###################################################################


###################################################################
# Methods
###################################################################

# display the help message

sub display_help {
  print "annotations_update.pl -help -oldseq [old sequence file] -newseq [new sequence file] -tab [annotation file] -speedup [true|false] -filtering [blast score]\n\n";
  print "\t-help a help message\n";
  print "\t-tab the annotation file (tab or embl file)\n";
  print "\t-oldseq the previous sequence file (embl or fasta format)\n";
  print "\t-newseq the new sequence file (embl or fasta format)\n";
  print "\t-speedup blast search speedup - speedup is activated per default\n";
  print "\t-filtering give the blast score limit below the feature won't be mapped  - no filtering per default\n";
}


###################################################################
# get min and max from a set of ranges
# the purpose is to blast against a big range to make sure the mapping will be correct
# so if the gene contains several exons, 
# the query sequence will be the all gene sequence - exons + introns.
###################################################################

sub create_range {
  my (@ranges) = @_;

  my $strand;
  my $start = 10**12;
  my $stop  = 0;

  foreach my $range (@ranges) {
    if ($range->start < $start) {
      $start = $range->start;
    }
    if ($range->end > $stop) {
      $stop = $range->end;
    }
    $strand = $range->strand;
  }
  my $new_range = Bio::PSU::Range->new(-start       => $start,
				       -end         => $stop,
				       -strand      => $strand);
  return $new_range;
}


######################################################################
# get the subsequence string from a seq object, giving a range object
# take always the forward sequence, because the mapping is always done on the forward strand
######################################################################

sub getSubSequence {
  my ($range,  $oldseq_obj) = @_;

  my $sub_seq = $oldseq_obj->subseq ($range->start, $range->end, 0);
  return $sub_seq->str;
}


####################################################################
# run a blast search and return the results as a result PSU object
# use ONLY WuBLAST 1.4 !!
##################################################################

sub doBlast {
  my ($curdir, $querySequence, $subjectSequence, $speedup, $is_exon_mapping) = @_;
  
  my $blast_dir = "blast_search";
  my $blast_output_name = $blast_dir . "/blast.res";
  my $blast_output_name_path = $curdir . "/" . $blast_output_name;
  
  my $queryFile   = $blast_dir . "/query.fasta";
  my $subjectFile = $blast_dir . "/subject.fasta";
  open QUERY,   ">$queryFile"   or die "can't open $queryFile\n";
  open SUBJECT, ">$subjectFile" or die "can't open $subjectFile\n";

  print QUERY ">query sequence\n";
  my $i = 0;
  my $length = 60;
  while ($i < length ($querySequence)) {
    print QUERY (substr ($querySequence, $i, $length) . "\n");
    $i = $i + $length;
  }
  print SUBJECT ">subject sequence\n";
  $i = 0;
  while ($i < length ($subjectSequence)) {
    print SUBJECT (substr ($subjectSequence, $i, $length) . "\n");
    $i = $i + $length;
  }

  close QUERY;
  close SUBJECT;

  # indexing the subject database

  qx/pressdb $curdir\/$subjectFile/;

  print STDERR "\tblast search running ...\n";
  
  my $subjectFile_path = $curdir . "/" . $subjectFile;
  my $queryFile_path   = $curdir . "/" . $queryFile;

  my $blast_options = "-warnings -hspmax 20 ";
  if ($speedup =~ /true/i) {
    $blast_options = $blast_options . "-top -span ";
  }
  if ($is_exon_mapping) {
    $blast_options = $blast_options . "W=8";
  }
  else {
    $blast_options = $blast_options . "W=20";
  }

  print STDERR "\tblast_options: $blast_options\n";

  @args = ("blastn $subjectFile_path $queryFile_path $blast_options > $blast_output_name_path");
  system (@args);
  
  my $bfh = Bio::PSU::IO::BufferFH->new(-file => $blast_output_name_path);
  my $search = Bio::PSU::IO::Blast::Search->new(-bfh => $bfh);
  my $result = $search->next_result;
  
  if ($debug && not defined $result) {
    print STDERR "no search result!\n";
  }

  return $result;
}


####################################################################
# select the best HSP and return it
####################################################################

# HSP : High Scoring Pair

sub getBestHSP {
  my ($result, $hit_filtering) = @_;
  my $best_hsp = undef;
  my $score = 0;

  while (defined (my $hit = $result->next_hit)) {
    if ($debug) {
      print STDERR "at least one hit ...\n";
    }
    while (defined (my $hsp = $hit->next_hsp)) {
      if ($hsp->score > $score) {
	$score = $hsp->score;
	$best_hsp = $hsp;
      }
    }
  }
  if (not defined $best_hsp) {
    print STDERR "\tno hits found\n";
    return undef;
  }
  else {
    print STDERR "\thit found whose score is : " . $best_hsp->score . "\n";
    if ($best_hsp->score > $hit_filtering) {
      return $best_hsp;
    }
    else {
      print STDERR "\tno significant hit found\n";
      return undef;
    }
  }
}


####################################################################
# create a new range object from a HSP object
####################################################################

sub getRangeBestHSP {
  my ($best_hsp) = @_;

  my $start  = $best_hsp->s_begin;
  my $end    = $best_hsp->s_end;
  my $strand = $best_hsp->s_strand;

  my $range = Bio::PSU::Range->new(-start       => $start,
                                   -end         => $end,
                                   -strand      => $strand);
  return $range;
}


####################################################################
# map the coordinates of a feature from a sequence to another one
#
# return a range object
# if can't find it, return an undefined range object
####################################################################

sub getMappedRange {
  my ($curdir, $querySequence, $subjectSequence, $speedup, $hit_filtering, $is_exon_mapping) = @_;

  # mapping algo :
  # => exact matching
  # if not found, use blast
  
  my $range = exactMatching (lc ($querySequence), lc ($subjectSequence));

  # if query sequence or subject sequence is less than 8 bases, no blast search

  my $canBlast = "false";
  my $word_length = 20;
  if ($is_exon_mapping) {
    $word_length = 8;
  }
  if ((length ($querySequence) > $word_length) && (length ($subjectSequence) > $word_length)) {
    $canBlast = "true";
  }
  else {
    print STDERR "\tcan't blast querySequence, $querySequence.\n";
  }

  if (($canBlast =~ /true/i) && (not (defined $range))) {
    #############
    # run blast
    ##
    my $result     = doBlast ($curdir, $querySequence, $subjectSequence, $speedup, $is_exon_mapping);
    #############
    # blast parsing to get the new subject subsequence
    ##
    my $best_hsp   = getBestHSP ($result, $hit_filtering);
    if (defined $best_hsp) {
      #############
      # get a new range
      ##
      $range = getRangeBestHSP ($best_hsp);
      if ($debug) {
	print STDERR "range for new subject sequence : " . $range->start . ", " . $range->end . "\n";
      }
    }
  }
  
  return $range;
}


####################################################################
# match the presence of the query sequence within the subject sequence
# return a range object
# if can't find it, return an undefined range object
####################################################################

sub exactMatching {
  my ($querySequence, $subjectSequence) = @_;

  my $range = undef;
  my $start;

  my $offset = 0;
  if (($start = index ($subjectSequence, $querySequence, $offset)) > -1) {
    # coordinates origin in EMBL in 1, not 0.
    $start++;
    print STDERR "\texact matching found\n";
    my $end = $start + length ($querySequence) - 1;
    $range = Bio::PSU::Range->new(-start  => $start,
				  -end    => $end,
				  -strand => 1);
  }
  return $range;
}


sub check_gene_accurracy {
  my ($feature, $newseq_obj) = @_;
  my @notes = ();

  # would get confused if cds with exons on different strands

  my $i = 1;
  my $nb_ranges = $feature->ranges+0;
  if ($feature->has_ranges) {
    foreach my $range ($feature->ranges) {
      
      my $sequence = $newseq_obj->str;

      # check on the appropriate strand

      # feature on the forward strand

      if ($range->strand == 1) {
	# check if the first exon starts with a start codon
	if ($debug) {
	  print STDERR "checking on the forward strand...\n";
	}
	if ($i == 1) {
	  my $first_codon = substr ($sequence, $range->start-1, 3);
	  if ($debug) {
	    print STDERR "first codon : $first_codon - (should be atg)\n";
	  }
	  if (not (lc($first_codon) =~ /atg/)) {
	    my $note = "first exon doesn't start with a start codon";
	    push (@notes, $note);
	  }
	}
	# check if the last exon ends with a stop codon
	if ($i == $nb_ranges) {
	  my $stop_codon = substr ($sequence, $range->end-3, 3);
	  if ($debug) {
	    print STDERR "stop codon : $stop_codon - (should be taa or tag or tga)\n";
	  }
	    if (not (lc($stop_codon) =~ /taa|tag|tga/)) {
	      my $note = "last exon doesn't end with a stop codon";
	      push (@notes, $note);
	    }
	}
      }
      # feature on the reverse strand
      else {
	# check if the first exon starts with a start codon
	if ($debug) {
	  print STDERR "checking on the reverse strand...\n";
	}
	if ($i == $nb_ranges) {
	  my $first_codon = substr ($sequence, $range->end-3, 3);
	  if ($debug) {
	    print STDERR "first codon : $first_codon - (should display cat)\n";
	  }
	  if (not (lc($first_codon) =~ /cat/)) {
	    my $note = "first exon doesn't start with a start codon";
	    push (@notes, $note);
	  }
	}
	# check if the last exon ends with a stop codon
	if ($i == 1) {
	  my $stop_codon = substr ($sequence, $range->start-1, 3);
	  if ($debug) {
	    print STDERR "stop codon : $stop_codon - (should display tta or cta or tca)\n";
	  }
	  if (not (lc ($stop_codon) =~ /tta|cta|tca/)) {
	    my $note = "last exon doesn't end with a stop codon";
	    push (@notes, $note);
	  }
	}
      }
      $i++;
    }
  }  

  if ((@notes+0) > 0) {
    $feature->qadd('note', @notes);
  }

  return $feature;
}

# take as input a feature and map it into the new sequence
# return a feature with a new range object
# if feature without range, return the given feature without any processing
# if failed, return undef

sub mapping {
  my ($feature, $oldseq_obj, $newseq_obj, $curdir, $speedup, $filtering) = @_;

  if ($feature->has_ranges) {
    my @ranges = $feature->ranges;
    my $nb_ranges = @ranges;

    # mapping of the feature
    
    # create a range which will be converted into the query sequence, by taking the min and the max corrdinates
    # min will the start position
    # max will be the end position
    my $range  = create_range (@ranges);
    my $querySequence   = getSubSequence ($range, $oldseq_obj);
    my $subjectSequence = $newseq_obj->str;

    # mapping of the subfeatures
    # filtering on the mapping of the gene location => e.g $filtering = 300 - per default 0
    # is_exon_mapping => 0
    $range = getMappedRange ($curdir, $querySequence, $subjectSequence, $speedup, $filtering, 0);
    my $failed = "false";
    my @new_ranges = ();
    
    if (not defined $range) {
      if ($debug) {
	print STDERR "failed - $range not defined\n";
      }
      $failed = "true";
    }
    elsif ($nb_ranges <= 1) {
      # if only one range, the mapping for this feature is done
      if ($debug) {
	print STDERR "only one range - no submapping\n";
      }
      $range->strand($ranges[0]->strand);
      push (@new_ranges, $range);
    }
    else {
      print STDERR "submapping...\n";
      if ($debug) {
	print STDERR "working on the following range for this feature: start, end " . $range->start . ", " . $range->end . "\n"; 
      }
      
      my $subjectSequence = getSubSequence ($range, $newseq_obj);
      
      # foreach subfeature for the current feature (e.g.) a set of exons, map into the new seq.
      # if one of the subfeature can not be mapped, copy the current feature into a failure tab file.
      
      $failed = "false";

      foreach my $sub_range ($feature->ranges) {
	if ($debug) {
	  print STDERR "mapping new sub feature mapping...\n";
	}
	
	my $strand = $sub_range->strand;
	my $querySequence = getSubSequence ($sub_range, $oldseq_obj);
	
	# create the new range
	# no filtering of the blast hits on the exons mapping => $filtering = 0
	# is_exon_mapping => 1
	my $tmp_range = getMappedRange ($curdir, $querySequence, $subjectSequence, $speedup, 0, 1);
	# if found a range
	# position it into the whole new sequence
	if (defined $tmp_range) {
	  
	  my $new_sub_range = Bio::PSU::Range->new(-start  => $range->start+$tmp_range->start-1,
						   -end    => $range->start+$tmp_range->end-1,
						   -strand => $strand);
	  if ($debug) {
	    print STDERR "found new subrange start, strand : " . $new_sub_range->start . ", " . $new_sub_range->strand  . "\n";
	  }
	  push (@new_ranges, $new_sub_range);
	  if ($debug) {
	    print STDERR "nb ranges after insert: " . (@new_ranges+0) . "\n";
	  }
	}
	else {
	  if ($debug) {
	    print STDERR "submapping failed!\n";
	  }
	  $failed = "true";
	  # add the subject range found into a note within $feature
	  if ($strand == 1) {
	    $feature->qadd('note', "failed but should be on the forward strand, starting at " . $range->start . " and ending at " . $range->end);
	  }
	  else {
	    $feature->qadd('note', "failed but should be on the reverse strand, starting at " . $range->start . " and ending at " . $range->end);
	  }
	  last;
	}
      }
    }
    if ($failed =~ /true/i) {
      return undef;
    }
    else {
      my $new_feature = $feature->clone(-ranges => \@new_ranges);
      
      if ($debug) {
	print STDERR "Feature range:" . Dumper (@new_ranges) . "\n";
      }
      
      # check if it makes sense
      # ie if fisrt exon has a start codon
      # and the last exon the stop codon
      # if not accurate, add a note message
      
      if ($new_feature->key =~ /cds/i) {
	$new_feature = check_gene_accurracy ($new_feature, $newseq_obj);
      }

      return $new_feature;
    }
  }
  else {
    # no range, return without any processing the given feature
    return $feature;
  }
}
