#!/usr/local/bin/perl5.6.0 -w

use strict;
use Getopt::Std;
use File::Basename;
use POSIX;
use Bio::PSU::SearchFactory;

# purpose: This script will chop up a large sequence, run blast on each bit and
# purpose: then write out an EMBL feature table and a MSPcrunch -d file
# purpose: containing the hits.

my $piece_size = 50000;
my $min_score = 0;
my $min_percent_id = 0;

# show a usage message and then exit
sub usage
{
  die <<EOF;
usage: $0 [-l] [-t|-x|-tx] [-S size] [-j num] [-M num] [-u path] database input_file [blast options]
   or: $0 [-M num] blast_output_file
   or: $0 -h (for help)

  -b      use bsub to share the blast jobs over multiple machines
  -l      use lsrun to share the blast jobs (not recommended)
  -x      selects blastx instead of blastn
  -t      selects tblastn instead of blastn
  -tx     selects tblastx instead of blastn
  -d      delete the blast output as soon as possible after processing
  -j num  sets the number of blast jobs to run simultaneously
  -M num  sets the minimum score for the output hits (default $min_score)
  -P num  sets the minimum percent identity for the output hits (default $min_percent_id)
  -S num  sets the maximum size of sequences sent to blast (default $piece_size)
  -w      use wublast instead of blast2
  -2      use blast2 not blast1.4 (this is the default)
  -1      use blast1.4 not blast2

The first form of the command will run blast on the given input_file
against the given database.  The -S flag sets the size of the pieces of
sequence that are passed to blast.  All arguments after the input_file
parameter will be passed directly to blast.

The second form reads and processes the given blast_output_file to
produce an MSPcrunch crunch file and an EMBL feature table.

In both cases there are two output files.  One is an EMBL feature
table with one feature per hit.  The other is a file in the
format generated by MSPcrunch -d.
EOF
}

my $blast2_program = "blastall_2.0.11";

my $blast_2_2_2 = "/usr/local/pubseq/bin/ncbi_blast/blastall";

if (-x $blast_2_2_2) {
  $blast2_program = $blast_2_2_2;
} else {
  warn "$0: can't find blast version 2.2.2: defailing to version 2.0.11\n";
}

my $blast_program  = "blastn";
my $blast_1_suffix = "_1.4";

# Sanger Centre only
my $blast_wrapper = "/nfs/pathsoft/psu/genexec/perl/src/blastwrap.pl";

my $lsrun          = "lsrun";
my $bsub           = "bsub";
my $batch_queue    = "babelq2";
my $child_count    = 1;

if (!@ARGV) {
  warn "error: not enough arguments\n\n";
  usage;
}

if ($ARGV[0] eq "-h") {
  usage;
}

my $version_flag = "";

my $blast_version      = "2";
my $use_lsrun          = 0;
my $use_bsub           = 0;
my $delete_blast_files = 0;

my %options = ();

getopts ("dblw21xtj:M:P:S:", \%options);

if ((exists $options{2}) + (exists $options{w})+ (exists $options{1}) > 1) {
  die "$0: error the -1, -2 and -w flags are mutually exclusive\n";
}

if (exists $options{j}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  $child_count = $options{j};
}

my $filter_flags = "";

if (exists $options{M}) {
  $min_score = $options{M};
  $filter_flags .= "_M${min_score}";
}

if (exists $options{P}) {
  $min_percent_id = $options{P};
  $filter_flags .= "_P${min_percent_id}";
}

if (exists $options{S}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  $piece_size = $options{S};
}

if (exists $options{1}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  $blast_program .= $blast_1_suffix;
  $blast_version = "1";
}

if (exists $options{2}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  $blast_version = "2";
}

if (exists $options{w}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  $blast_program = 'wu' . $blast_program;
  $blast_version = "w";
}

if (exists $options{l}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  warn "warning: the -l flag is not recommended - use -b instead\n";

  $use_lsrun = 1;
}

if (exists $options{b}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  if ($use_lsrun) {
    warn "ignoring -l flag\n";
    $use_lsrun = 0;
  }

  $use_bsub = 1;
}

if (exists $options{d}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  $delete_blast_files = 1;
}

if (exists $options{'x'}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  $blast_program =~ s/blastn/blastx/;
  $version_flag .= "x_";
}

if (exists $options{t}) {
  if (@ARGV < 2) {
    warn "error: not enough arguments\n\n";
    usage;
  }

  $blast_program = 't' . $blast_program;
  $version_flag .= "t_";
}

if (@ARGV == 1) {
  unlink "$ARGV[0].crunch", "$ARGV[0].tab";

  process_blast ($ARGV[0], 0);

  print STDERR "wrote $ARGV[0].crunch and $ARGV[0].tab\n";
  exit;
}

if (@ARGV < 1) {
  warn "error: not enough arguments\n\n";
  usage;
}

my @blast_options;
#my @blast_options = qw { -span1 V=2000 B=2000 };

if ($blast_version eq "2") {
  @blast_options = ("-p", $blast_program);
  $version_flag .= "2";
} else {
  if ($blast_version eq "w") {
    @blast_options = qw { B=1000000 -span V=1000000 -H=0 -hspmax 999999 };
    $version_flag .= "w";
  } else {
    @blast_options = qw { B=1000000 -span V=1000000 -H=0 -hspmax 999999 };
    $version_flag .= "1";
  }
}


my $database = shift;
my $in_file  = shift;

if (!open (IN, "$in_file")) {
  if (!open (IN, "$in_file.seq")) {
    die "can't open $in_file or $in_file.seq\n";
  } else {
    print STDERR "note: opened $in_file.seq\n";
  }
}

my $database_name = basename($database);
my $base_name     = "big_blast_" . $version_flag . "_" . $piece_size .
    "_${database_name}_$in_file";
my $base_name_for_filtered_output = "big_blast_" . $version_flag .
    "_" . $filter_flags . $piece_size . "_${database_name}_$in_file";

push @blast_options, @ARGV;

# the files that contain the small pieces of sequence
my @piece_files = ();
my @piece_offsets = ();

# make a fasta sequence file for the given piece of sequence
# args: $sequence - a string containing the sequence to write
#       $offset - the position in the original sequence that the sequence was
#                 taken from
sub make_piece_file
{
  my ($sequence, $offset) = ($_[0], $_[1] + 1);
  my $padded   = sprintf ("%8.8d", $offset);
  my $out_file = "$base_name.$padded";

  $sequence =~ s/(.{60})/$1\n/g;

  if (-s $out_file) {
    print STDERR "not writing $out_file - file exists\n";
    return $out_file;
  }

  open OUT, ">$out_file" or die "$0: can't open $out_file\n";

  print OUT ">$out_file\n";
  print OUT $sequence,"\n";

  close OUT;

  $out_file;
}

my $sequence = "";

while (<IN>) {
  if (!/^>/) {
    chomp;
    $sequence .= $_;
  }
}

my $bases_so_far = 0;

while ($bases_so_far < length $sequence) {
  my $piece = substr $sequence, $bases_so_far, $piece_size;
  push @piece_files, (make_piece_file $piece, $bases_so_far);
  push @piece_offsets, $bases_so_far;

  $bases_so_far += length $piece;
}

print STDERR "sequence split into ", scalar (@piece_files), " pieces\n";

my @children = ();

# returns true if the given file seems like a valid blast output file
sub check_blast_output
{
  my ($in_file) = @_;

  if (!open IN_FILE, $in_file) {
    return 0;
  }

  local $/ = undef;

  my $file_size = -s IN_FILE;

  if (!$file_size) {
    close IN_FILE;
    print STDERR "$in_file exists but is truncated\n";
    return 0;
  }

  my $tail_size = 10000;

  if ($file_size > $tail_size) {
    if (!seek IN_FILE, $file_size - $tail_size, SEEK_SET) {
      warn "can't seek on $in_file\n";
      close IN_FILE;
      return 0;
    }
  }

  my $end_of_file = <IN_FILE>;

  if ($end_of_file =~ m/effective length of database|Total cpu time/) {
    close IN_FILE;
    return 1;
  } else {
    close IN_FILE;
    print STDERR "$in_file exists but is truncated\n";
    return 0;
  }
}


# process the results of a blast job
# args: $file - the blast output file to process
#       $offset - the position in the original sequence that this blast job
#                 covers
sub process_blast
{
  my ($file, $offset) = @_;

  # wait for (up-to) 20 secs for NFS to catch up
  for (1..20) {
    if (-s $file) {
      last;
    } else {
      sleep 1;
    }
  }

  my $crunch_file = "$file.crunch" . $filter_flags;
  my $tab_file    = "$file.tab" . $filter_flags;

  if (-s $crunch_file && -s $tab_file) {
    print STDERR "skipping $crunch_file and $tab_file - file exists\n";
    return;
  }

  open CRUNCH_OUT, ">$crunch_file" or die "couldn't open $crunch_file\n";
  open TAB_OUT,    ">$tab_file"    or die "couldn't open $tab_file\n";

  print STDERR "opened $crunch_file for writing\n";

  my $blast_obj;

  eval {
    $blast_obj = Bio::PSU::SearchFactory->make(-file     => "$file",
                                               -program  => "blast");
  };

  if ($@) {
    print STDERR $@;
    return;
  }

  my $results = 0;

  while (defined (my $result = $blast_obj->next_result)) {
    $results++;
    while (defined (my $hit = $result->next_hit)) {
      my $s_desc = $hit->s_desc;
      my $s_name = $hit->s_id;

      my $q_name = $hit->q_id;

      if (!defined $s_name) {
        $s_name = "unknown"
      }

      $q_name =~ s/(\S+).*/$1/;

      # don't put / in labels
      $s_name =~ s!/!-!g;

      my $hsp;
      while (defined ($hsp = $hit->next_hsp)) {
        # MSPcrunch output:
        # 154 55.87 35135 35449     NoName 1958138 1958445 TB.dna

        my $q_start    = $hsp->q_begin + $offset;
        my $q_end      = $hsp->q_end + $offset;
        my $s_start    = $hsp->s_begin;
        my $s_end      = $hsp->s_end;

        my $percent_id = $hsp->percent;

        my $score      = $hsp->score;

        next if ($score < $min_score);
        next if ($percent_id < $min_percent_id);

        if ($q_end < $q_start) {
          ($q_end, $q_start) = ($q_start, $q_end)
        }

        if ($s_end < $s_start) {
          ($s_end, $s_start) = ($s_start, $s_end)
        }

        printf CRUNCH_OUT "%d %3.2f %d %d $q_name %d %d $s_name $s_desc\n",
        $hsp->score,
        $percent_id,
        (do
         {
           if ($hsp->q_strand == -1) {
             ($q_end, $q_start)
           } else {
             ($q_start, $q_end)
           }
         }),
         (do
          {
            if (defined $hsp->s_strand and $hsp->s_strand == -1) {
              ($s_end, $s_start)
            } else {
              ($s_start, $s_end)
            }
          });

        if ($hsp->q_strand == -1 ^ $hsp->s_strand == -1) {
          print TAB_OUT "FT   \U$blast_program\E_HIT      complement($q_end..$q_start)\n";
        } else {
          print TAB_OUT "FT   \U$blast_program\E_HIT      $q_start..$q_end\n";
        }

        my $s_desc_for_tab_file = $s_desc;

        $s_desc_for_tab_file =~ s/\"/\"\"/g;

        print TAB_OUT "FT                   /note=\"match to $s_name $s_start..$s_end  blast score $score  percent identity $percent_id  $s_desc_for_tab_file\"\n";
        print TAB_OUT "FT                   /label=$s_name\n";
        print TAB_OUT "FT                   /score=$percent_id\n";

        my $colour = "255" . (" " . int((100 - $percent_id) * 2.55)) x 2;

        print TAB_OUT "FT                   /colour=$colour\n";
      }
    }
  }

  if ($delete_blast_files) {
    unlink $file;
  }

  close CRUNCH_OUT;
  close TAB_OUT;

  if (! $results) {
    warn "Couldn't read any blast results from $file\n";
    unlink $tab_file, $crunch_file;
  }
}

# start a child process (and blast job) and put the id of the child in
# @children
# args: $file - the sequence file to run blast on
#       $offset - the position in the original sequence that this file piece
#                 starts at
sub start_child ($$)
{
  my ($file, $offset) = @_;

  my $output_file = "$file.out";
  my $child_pid   = fork;

  if ($child_pid == 0) {
    # the child

    if (check_blast_output $output_file) {
      print STDERR "skipping $file - $output_file seems OK\n"
    } else {
      if (!$use_bsub) {
        open(SAVEOUT, ">&STDOUT");

        # to avoid a warning about SAVEOUT being used only once
        print SAVEOUT "";

        close STDOUT;
        open STDOUT, ">$output_file" or die "can't open $output_file\n";
      }

      my @exec_args = ();

      if ($blast_version eq "2") {
        @exec_args = ($blast2_program,
                      "-d", $database,
                      "-i", $file, @blast_options)
      } else {
        @exec_args = ($blast_program,
                      $database,
                      $file, @blast_options)
      }

      if (-x $blast_wrapper) {
        # Sanger Centre only
        unshift(@exec_args, $blast_wrapper, "-v");
      }

      if ($use_lsrun) {
        print STDERR "starting lsrun of: @exec_args\n";
        system $lsrun, "-v", @exec_args
      } else {
        if ($use_bsub) {
          my $tmp_error_file = "big_blast_tmp_errors_$$";

          # bsub sends it's output to STDOUT by default
          open STDOUT, ">&STDERR" or die "can't reopen STDOUT\n";

          print "starting bsub of: @exec_args\n";

          # errors will go to STDERR
          system ($bsub, "-q", $batch_queue,
                  "-I", "-o", $output_file, @exec_args);
          if ($?) {
            # wait for 20 secs for NFS to catch up
            LOOP: for (1..20) {
              if (-s $output_file) {
                open OUTPUT_FILE, $output_file or
                  die "can't open $output_file\n";
                while (my $line = <OUTPUT_FILE>) {
                  print STDERR $line
                }
                close OUTPUT_FILE;
                last LOOP;
              } else {
                sleep (1);
              }
            }
          }
        } else {
          print STDERR "starting @exec_args on localhost\n";
          system @exec_args;
        }
      }

      if (!$use_bsub) {
        close STDOUT;
        open(STDOUT, ">&SAVEOUT");
      }
    }

    process_blast $output_file, $offset;

    exit (0);
  } else {
    if ($child_pid == -1) {
      die "$0: failed to start child process\n"
    } else {
      # the parent
      push @children, $child_pid;

      print STDERR "started process with id $child_pid\n";

      if ($use_lsrun) {
        # sleep so that the machines aren't swamped by jobs
        if (length $sequence > 100000 and length $sequence <= 1000000) {
          sleep 2
        } else {
          if (length $sequence > 1000000) {
            sleep 3
          } else {
            sleep 1
          }
        }
      }
    }
  }
}

# returns when one of the child jobs finishes
sub wait_for_child
{
  my $child_pid = wait;

  @children = grep { $_ != $child_pid } @children;

  print STDERR "process with id $child_pid has finished\n";
}


for (my $i = 0 ; $i < @piece_files ; ++$i) {
  my $file = $piece_files[$i];

  if (@children < $child_count) {
    start_child $file, $piece_offsets[$i];
  } else {
    wait_for_child;
    start_child $file, $piece_offsets[$i];
  }
}

while (@children) {
  wait_for_child;
}


# wait for nfs
sleep 10 if $use_lsrun;

#  for (my $i = 0 ; $i < @piece_files ; ++$i) {
#    my $file = $piece_files[$i];

#    process_blast ($file . ".out", $piece_offsets[$i]);
#  }

my $crunch_file = $base_name . ".crunch" . $filter_flags;

open CRUNCH_OUT, ">$crunch_file";

for my $in_file (map {$_ . ".out.crunch" . $filter_flags} @piece_files) {
  open IN_FILE, $in_file or die "can't open $in_file\n";
  print CRUNCH_OUT <IN_FILE>;
  close IN_FILE;
}

close CRUNCH_OUT;

print STDERR "wrote MSPcrunch output to $crunch_file\n";

my $tab_file = $base_name . ".tab" . $filter_flags;

open TAB_OUT, ">$tab_file";

for my $in_file (map {$_ . ".out.tab" . $filter_flags} @piece_files) {
  open IN_FILE, $in_file or die "can't open $in_file\n";
  print TAB_OUT <IN_FILE>;
  close IN_FILE;
}

close TAB_OUT;

print STDERR "wrote tab file output to $tab_file\n";
