#!/nfs/newlocal/bin/perl5.6.0 -w

#
# this script to wrap an ESTs Clustering process
# Input:   one Fasta File with the all set of sequences
# Ouptput: a set of fasta format cluster files, an Alignment file (.aln) and a unique fasta file with a the contigs and singlets.
#

#
# e.g. perl -w est_clustering.pl -i [input_file] -o [output_directory] -c [config_file] -r true -e estate -m babel -a true -s true
#

# process : 
# 1/ Run RepeatMasker - optional => -repeatMasker true/false 
# 2/ Run a clustering tool EState or Phrap or cap3 or tigr => -estTool estate/phrap/cap3/tigr
# 3/ if Estate: run cons and clustalw to create the alignment file (.aln)
# NB: if phrap we can not do that directly. GAP4 can do that but we need to automatize on the all set of clusters
# 4/ Run a perl program who generate the unique fasta file with contigs and singlets information
# 5/ Run pressdb to index the output generated fasta file

# External Modules

# Constraint variables declaration before using them
use strict;
##########################################

use Data::Dumper;

BEGIN {
  my $_path="/nfs/disk222/yeastpub/scripts/perl/axk/";
  unshift (@INC, $_path);
}

# Module for managing global application configuration information
use ConfigReader::Simple;
##########################################

=head1 NAME

est_clustering.pl - a ESTs Clustering wrapper script

=head1 SYNOPSIS

  perl -w est_clustering.pl -help -i [input_file|directory_input_file] -o [output_directory] -c [config_file] -r [true|false] -e [estate|phrap|cap3|tigr] -m [machine] -a [true|false] -s [true|false]

  -help a help message
  -i the input
  -o the output
  -c the config file
  -r repeats masking process
  -e the ESTs clustering tool
  -m the machine where the process will be done
  -a generates the alignment files by running clustalw
  -s add the ESTs singletons into the output fasta file

=head1 DESCRIPTION

  this script to wrap an ESTs Clustering process

  @author Arnaud Kerhornou - axk@sanger.ac.uk

=cut

=head1 ALGORITHM

=cut

##########################################
=pod

=item * Configuration loading from the command line

=cut
##########################################

my $_help          = "-help";
my $_input         = "-i";
my $_output        = "-o";
my $_config        = "-c";
my $_repeatMasking = "-r";
my $_estTool       = "-e";
my $_machine       = "-m";
my $_align         = "-a";
my $_singlets      = "-singlets";

# no input per default
my $input_name     = "-1";

# The options per default

my $output_name    = "clusters";
my $config_name    = "/nfs/disk222/yeastpub/scripts/perl/axk/est_clustering.cfg";
my $repeat_masking = "false";
my $est_tool       = "phrap";
my $machine        = "babel";
my $align          = "false";
my $singlets       = "true";

my $cur_dir = qx/pwd/;
chomp ($cur_dir);

# Check if in the command line there is an input file name (-i option)
# the other options are not compulsory (set of options per default)
my $lgth = @ARGV + 0;
# if no args, display the help message
if ($lgth == 0) {
  print "specify at least an input file with the option \"-i\"\n";
  display_help ();
  exit 1;
}
elsif ($lgth >= 1) {
  my $i = 0;
  foreach my $element (@ARGV) {
    if ($element eq $_help) {
      display_help ();
      exit 1;
    }
    elsif ($element eq $_input) {
      $input_name = $ARGV[$i+1];
    }
    elsif ($element eq $_output) {
      $output_name = $ARGV[$i+1];
    }
    elsif ($element eq $_config) {
      $config_name = $ARGV[$i+1];
    }
    elsif ($element eq $_repeatMasking) {
      $repeat_masking = $ARGV[$i+1];
    }
    elsif ($element eq $_estTool) {
      $est_tool = $ARGV[$i+1];
    }
    elsif ($element eq $_machine) {
      $machine = $ARGV[$i+1];
    }
    elsif ($element eq $_align) {
      $align = $ARGV[$i+1];
    }
    elsif ($element eq $_singlets) {
      $singlets = $ARGV[$i+1];
    }
    # if at least one of the options is not recognized, display the help message and exit
    elsif ($element =~ /^-/) {
      print STDERR "option not recognized: $element\n";
      display_help ();
      exit 1;
    }
    $i ++;
  }
}
if ($input_name eq "-1") {
  print STDERR "ERROR - specify at least the name of an input file with the -i option\n";
  print STDERR "Syntax : perl est_clustering.pl -i [input_file]\n";
  exit 1;
}

#################################################
# INITIALISATION
#################################################

my $output_name_estate   = $output_name . ".final";

my $_singletons_name_file = undef;
my $_singletons_name_suffix = undef;
if (lc ($est_tool) eq "estate") {
  
  $_singletons_name_file = "singletons.fasta";
}
elsif (lc ($est_tool) eq "phrap") {
  $_singletons_name_suffix = ".singlets";
}
elsif (lc ($est_tool) eq "cap3") {
  $_singletons_name_suffix = ".cap.singlets";
}
elsif (lc ($est_tool) eq "tigr") {
  # ??????????????????????
  $_singletons_name_suffix = ".singlets";
}

my $phrap_version = "phrap.manyreads";
# my $phrap_version = "phrap";

my %ace_parsing_options;
if ($est_tool eq "phrap" && $phrap_version eq "phrap") {
  %ace_parsing_options = (
			  'ace_name_suffix'  => '.ace',
			  'contig_pattern'   => 'DNA\s(\S+)',
			  'AF_test'          => '^Assembled_from',
			  'AF_pattern'       => 'Assembled_from\*?\s+(\S+)\s+-?\d+\s+-?\d+',
			  'member_sequence'  => 'DNA\s(\S+)',
			 );
}
elsif ($est_tool eq "phrap" && $phrap_version eq "phrap.manyreads") {
  %ace_parsing_options = (
			  'ace_name_suffix' => '.ace',
			  'contig_pattern'  => 'CO\s(\S+)\s+.*',
			  'AF_test'         => '^AF',
			  'AF_pattern'      => 'AF\s+(\S+)\s+(C|U)\s+-?\d+.*',
			  'member_sequence' => 'RD\s+(\S+).*',
			 );
}
elsif ($est_tool eq "cap3") {
  %ace_parsing_options = (
			  'ace_name_suffix' => '.cap.ace',
			  'contig_pattern'  => 'CO\s(\S+)\s+.*',
			  'AF_test'         => '^AF',
			  'AF_pattern'      => 'AF\s+(\S+)\s+(C|U)\s+-?\d+.*',
			  'member_sequence' => 'RD\s+(\S+).*',
			 );
}
elsif ($est_tool eq "tigr") {
  %ace_parsing_options = (
			  'ace_name_suffix'  => '.ace',
			  'contig_pattern'   => 'DNA\s(\S+)',
			  'AF_test'          => '^Assembled_from',
			  'AF_pattern'       => 'Assembled_from\*?\s+(\S+)\s+-?\d+\s+-?\d+',
			  'member_sequence'  => 'DNA\s(\S+)',
			 );
}

my $input_name_base         = "$input_name.results";
my $output_name_base        = "$output_name";
my $output_name_estate_base = "$output_name_estate";

# remove previous output directories
  
my @args = "rm -rf $output_name_base*";
system (@args);
@args = "rm -rf $output_name_estate_base*";
system (@args);

##########################################
=pod

=item * Config File Loading

=cut
##########################################

# phrap parameters
  
my $_word        = "word";
my $_display     = "display";
my $_min_overlap = "min_overlap";
my $_node_space  = "node_space";

my $cfg = ConfigReader::Simple->new ("$config_name", ["$_word $_display $_min_overlap $_node_space"]) or die "No such file or directory";
$cfg->parse ();

##########################################
=pod

=item * Repeat Masking

=cut
  ##########################################
  
if (lc ($repeat_masking) eq "true") {
  $input_name = repeatMasking ($input_name);
}

##########################################
=pod

=item * ESTs clustering

=cut
##########################################

if (lc ($est_tool) eq "estate") {
  @args = "mkdir $output_name_estate";
  system (@args);
  runESTate ($cfg, $input_name, $output_name, $output_name_estate, $_word, $_display, $_min_overlap);
}
elsif (lc ($est_tool) eq "phrap") {
  @args = "mkdir $output_name";
  system (@args);
  runPhrap ($cfg, $input_name, $_node_space, $phrap_version);
}
elsif ($est_tool =~ /cap3/i) {
  @args = "mkdir $output_name";
  system (@args);
  runCap3 ($input_name);
}
elsif ($est_tool =~ /tigr/i) {
  @args = "mkdir $output_name";
  system (@args);
  runTigr ($input_name);
}
else {
  print STDERR "can not process the clustering with the following tool, $est_tool\n";
  exit 0;
}

my @a_info = ();
my @clusters_info = ();

# ace parsing

@clusters_info = getClustersInfoFromAce ($cur_dir, $input_name, 0, undef, %ace_parsing_options);

# generate the "clusters" directory with a set of cluster files

cluster_files_generation ($cur_dir, $output_name, @clusters_info);

##################################
# part 3
##################################

# 3/ Run the perl script to generate a unique fasta file with contigs and (optionally) singlets information

@a_info = gather_contig_info ($cur_dir, $output_name);

if ($singlets =~ /true/i) {
  if ($est_tool =~ /phrap|cap3|tigr/) {
    $_singletons_name_file = $input_name . $_singletons_name_suffix;
  }
  my $singletons_file_path = $cur_dir . "/" . $_singletons_name_file;  
  my @a_info_tmp = gather_singlets_info ($singletons_file_path, "true");
  @a_info = concat (\@a_info, \@a_info_tmp);
}

unique_fasta_file_generation ($cur_dir, $input_name_base, @a_info);

##########
# 4/ Alignment Files generation 
    # if phrap, parse the ace file to generate a clusters directory in order to generate the same output than ESTate
    # the purpose is to run clustalw on the phrap output

if (($est_tool =~ /phrap|cap3|tigr/i) && lc ($align) eq "true") {
  run_clustalw ($cur_dir, $output_name);
}

print "process done\n";

###################################################################
# The end
###################################################################

 # display the help message

sub display_help {
  print "perl -w est_clustering.pl -help -i [input_file] -o [output_directory] -c [config_file] -r [true|false] -e [estate|phrap] -m [machine] -a [true/false]\n\n";
  print "\t-help a help message\n";
  print "\t-i the input       - (no default, always specify it)\n";
  print "\t-o the output      - (default: \"clusters\" directory)\n";
  print "\t-c the config file - (default file: /nfs/disk222/yeastpub/scripts/perl/axk/est_clustering.cfg)\n";
  print "\t-r repeats masking process  - (no repeats masking process per default)\n";
  print "\t-e the ESTs clustering tool - (default: phrap)\n";
  print "\t-m the machine where the process will be done, via lsrun - (default: babel)\n";
  print "\t-a alignment process using clustalw - (default: false)\n";
  print "\t-singlets add the ESTs singletons sequences into the output fasta file - (default: true)\n\n";
}


sub repeatMasking {
  
  my ($input_name) = @_;
  
  print STDOUT "\nrepeats masking processing ...\n";
  print STDOUT "command: lsrun -v /bin/csh -c \"unlimit;ulimit -a;RepeatMasker $input_name\"\n";
  
  my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;RepeatMasker $input_name\"");
  system (@args);
  
  # the input is now $input concatenated with ".masked"
  $input_name = $input_name . ".masked";
  
  print STDOUT "repeats masking done\n";
  print STDOUT "the input file is now $input_name\n\n";

  return $input_name;
}


sub runESTate {
  
  my ($cfg, $input_name, $output_name, $output_name_estate, $_word, $_display, $_min_overlap) = @_;
  
  print STDOUT "ESTs clustering processing using ESTate ...\n";
  my $input_name_ava = $input_name . ".ava";
  my $input_name_avarc = $input_name . ".avarc";
  my $word = $cfg->get($_word);
  
  print STDOUT "command: lsrun -v precluster -d $input_name -A $input_name_ava -R $input_name_avarc -w $word\n";
  @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit;/nfs/disk222/yeastpub/bio-soft/estate/bin/precluster -d $input_name -A $input_name_ava -R $input_name_avarc -w $word\"");
  # @args = ("unlimit;ulimit;/nfs/disk222/yeastpub/bio-soft/estate/bin/precluster -d $input_name -A $input_name_ava -R $input_name_avarc -w $word");
  system (@args);
  
  # display ON/OFF
  my $display = $cfg->get($_display);
  # minimum overlap alignment score
  my $min_overlap = $cfg->get($_min_overlap);
  
  print STDOUT "command: lsrun -v estcluster -d $input_name -A $input_name_ava -R $input_name_avarc -o $output_name -t $display -j $min_overlap\n";
  @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;/nfs/disk222/yeastpub/bio-soft/estate/bin/estcluster -d $input_name -A $input_name_ava -R $input_name_avarc -o $output_name -t $display -j $min_overlap\"");
  # @args = ("unlimit;ulimit -a;/nfs/disk222/yeastpub/bio-soft/estate/bin/estcluster -d $input_name -A $input_name_ava -R $input_name_avarc -o $output_name -t $display -j $min_overlap");
  system (@args);
  
  print STDOUT "ESTs clustering using ESTate is done\n";
  
}


sub runPhrap {
  
  my ($cfg, $input_name, $_node_space, $phrap_version) = @_;
  
  print STDOUT "ESTs clustering processing using Phrap ...\n";
  
  my $node_space = $cfg->get($_node_space);

  # if the database is pretty big, you can use phrap.manyreads
  
  if ($phrap_version eq "phrap") {
    my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;phrap $input_name -ace -node_space $node_space \"");
    system (@args);
  }
  else {
    my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;phrap.manyreads $input_name -ace -node_space $node_space \"");
    system (@args);
  }
  
  print STDOUT "\nESTs clustering using Phrap is done\n";
  
}


sub runCap3 {
  my ($input_name) = @_;
  
  print STDOUT "ESTs clustering processing using cap3 ...\n";
  
  my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a;cap3 $input_name >! $input_name.align \"");
  system (@args);
}


sub runTigr {
  my ($input_name) = @_;
  
  print STDOUT "ESTs clustering processing using Tigr ...\n";
  
  my @args = ("lsrun -v /bin/csh -c \"unlimit;ulimit -a; /nfs/pathsoft/external/bio-soft/TIGR_Assembler_v2/bin/run_TA_est $input_name\"");
  system (@args);
}


sub getClustersInfoFromAce {
  my ($cur_dir, $input_name, $is_reclustering, $previous_clusters_info_ref, %ace_parsing_options) = @_;

  my @previous_clusters_info = ();

  if (defined $previous_clusters_info_ref) {
    @previous_clusters_info = @{$previous_clusters_info_ref};
    # print STDERR "in getFromAce - previously " . @previous_clusters_info . " clusters\n";
  }

  my @clusters_info_tmp = parse_ace_file ($cur_dir, $input_name, %ace_parsing_options);

  if (not $is_reclustering) {
    return @clusters_info_tmp;
  }
  else {

    my @clusters_info = ();

    foreach my $cluster_ref (@clusters_info_tmp) {
      my %cluster = %$cluster_ref;
      my $previous_members_list_ref = $cluster{members_info};
      my @members_list = ();
      foreach my $member_ref (@$previous_members_list_ref) {
	my %member_info = %$member_ref;
	my $member = $member_info{line_info};

	# print STDERR "member, $member, mappping...\n";
	
	if ($member =~ /contig/i) {
	  my @members_list_tmp = map_cluster2members ($member, @previous_clusters_info);

	  # print STDERR "mapping results - members: @members_list_tmp\n";

	  push (@members_list, @members_list_tmp);
	}
	else {
	  push (@members_list, $member);
	}
      }
      my %new_cluster = (
			 'cluster_info' => $cluster{cluster_info},
			 'members_info' => \@members_list,
			);
      push (@clusters_info, \%new_cluster);
    }
    return @clusters_info;
  }
}


sub map_cluster2members {
  my ($member, @previous_clusters_info) = @_;

  # print STDERR "in map - previously " . @previous_clusters_info . " clusters\n";

  foreach my $cluster_ref (@previous_clusters_info) {
    my %cluster = %$cluster_ref;
    my %contig_info = %{$cluster{cluster_info}};
    my $description  = $contig_info{line_info};

    # print STDERR "description: $description\n";

    if ($description =~ /$member/i) {

      # print STDERR "description match found for contig $member\n";

      my @members_list_info = @{$cluster{members_info}};
      return @members_list_info;
    }
  }

  print STDERR "ERROR no member found for this contig, $member\n";
  return ();

}


sub parse_ace_file {
  
  my ($cur_dir, $input_name, %ace_parsing_options) = @_;
  
  my $ace_name = "$input_name" . $ace_parsing_options{ace_name_suffix};

  print STDERR "parsing the ace file, $ace_name...\n";
  
  # open the ace file
  
  open (ACE, "< $ace_name") or die "can't open the ace file, $ace_name !!!\n";
  
  # parse it
  
  # this data structure stores everything !!!!
  # this array will be returned to main
  my @clusters_list = ();
  
  while (<ACE>) {
    if ($_ =~ /$ace_parsing_options{contig_pattern}/) {
      my $contig_name = $1;
      # print STDERR "new cluster, contig name : $contig_name\n";
      # retrieve Contig sequence
      my $sequence_contig = "";
      # print STDERR "parsing contig sequence...\n";
      my $new_line = <ACE>;
      while (lc ($new_line) =~ /(a|c|t|g|n)\S+/) {
	chomp ($new_line);
	$sequence_contig = $sequence_contig . $new_line;
	$new_line = <ACE>;
      }
      # print STDERR "sequence_contig : $sequence_contig\n";

      $sequence_contig =~ s/\*//g;
      
      # then Base quality
      while (not $new_line =~ /$ace_parsing_options{AF_test}/) {
	$new_line = <ACE>;
      }
      my @members = ();
      # then information about the number of members within the current cluster
      while ($new_line =~ /$ace_parsing_options{AF_test}/) {
	$new_line =~ /$ace_parsing_options{AF_pattern}/;
	my $member_name = $1;
	# print STDERR "found member $member_name, members list : , new_line : $new_line\n";
	if (lc (is_in ($member_name, \@members)) eq "false") {
	  push (@members, $member_name);
	}
	$new_line = <ACE>;
      }
      my $members_nb = (@members+0);
      my $cpt = 0;
      my @members_list = ();

      # print STDERR "found $members_nb members for the current cluster\n";

      # Retrieve the name and the sequence of each member
      
      while ($cpt < $members_nb) {
	if ($new_line =~ /$ace_parsing_options{member_sequence}/) {
	  my $member_name = $1;
	  # print STDERR "member name : $member_name\n";
	  my $new_line = <ACE>;
	  my $sequence = "";
	  while (lc ($new_line) =~ /(a|c|t|g|n)\S+/) {
	    chomp ($new_line);
	    $sequence = $sequence . $new_line;
	    $new_line = <ACE>;
	  }
	  # print STDERR "member sequence : $sequence\n";
	  
	  # the sequence may contain * which don't fit with the fasta format, so remove them from the sequence
	  # this * is for a shift in the alignment

	  $sequence =~ s/\*//g;

	  my %h_member = (line_info => "$member_name",
			  sequence  => "$sequence",
			 );
	  $cpt ++;
	  push (@members_list, \%h_member);
	}
	$new_line = <ACE>;
      }

      my $line_info = $contig_name . " of $members_nb members : " . join ( ", ", @members);

      my %h_contig = (line_info => "$line_info",
		      sequence  => "$sequence_contig",
		     );

      # this data structure stores all the information about one contig and its members
      my %cluster_info = (cluster_info => \%h_contig,
			  members_info => \@members_list,
			 );
      push (@clusters_list, \%cluster_info);
      
      # print STDERR "Contig done - current line : $new_line\n";
      
    }
   
  }
  close (ACE);
  print STDERR "parsing $ace_name done - " . (@clusters_list+0) . " clusters found\n";

  return @clusters_list;
}


sub cluster_files_generation {

  my ($cur_dir, $output_name, @clusters_list) = @_;
  my $output_name_path = $cur_dir . "/" . $output_name;

  print STDERR "generating " . (@clusters_list+0) . " cluster files\n";

  foreach my $cluster_info_ref (@clusters_list) {
    my %cluster_info = %$cluster_info_ref;
    my $h_contig_ref = $cluster_info{cluster_info};
    my %h_contig = %$h_contig_ref;
    my $members_list_ref = $cluster_info{members_info};
    my @members_list = @$members_list_ref;
    my $line_info = $h_contig{line_info};
    $line_info =~ /Contig(\d+).+/;
    my $contig_index = $1;
    my $cluster_file_name = "cluster" . $contig_index . ".fasta";
    my $cluster_file_name_path = $output_name_path . "/" . $cluster_file_name;
    
    open (CLUSTER, "> $cluster_file_name_path");

    # write members information

    foreach my $h_member_ref (@members_list) {
      my %h_member = %$h_member_ref;
      print CLUSTER ">$h_member{line_info}\n";
      
      my $sequence = $h_member{sequence};
      $sequence = uc ($sequence);
      my $lgth = length ($sequence);
      my $i = 0;
      my $size = 60;
      while ($i < $lgth) {
	print CLUSTER substr ($sequence, $i, $size);
	print CLUSTER "\n";
	$i = $i + $size;
      }
    }

    # write contig information
    print CLUSTER ">$h_contig{line_info}\n";
    
    my $sequence = $h_contig{sequence};
    $sequence = uc ($sequence);
    my $lgth = length ($sequence);
    my $i = 0;
    my $size = 60;
    while ($i < $lgth) {
      print CLUSTER substr ($sequence, $i, $size);
      print CLUSTER "\n";
      $i = $i + $size;
    }
    close (CLUSTER);
  }
}


sub run_clustalw {
  
  my ($cur_dir, $output_name) = @_;
    
  print "running clustalw on each contig to generate the alignment file...\n";

  my $output_name_path = $cur_dir . "/" . $output_name;
  # print "output_path: $output_name_path.\n";
  
  # output directory files
  opendir (THISDIR, $output_name_path) or die "can not read this directory $output_name_path ...";
  my @files = grep {$_ =~ /^cluster.*fasta/} readdir THISDIR;

  # print STDOUT "files: \n@files\n";

  closedir THISDIR;
  
  foreach my $file_name (@files) {
    my @tmp = split (/\./, $file_name);
    my $cluster_name = $tmp[0];
    # print "cl name: $cluster_name\n";
    my $cluster_number = substr ($cluster_name, 7, length ($cluster_name)) + 0;
    # print STDOUT "cluster_number: $cluster_number\n";
    my $aln_file_name = $cluster_name . ".aln";
    my $contig_name = "CONTIG" . $cluster_number;
    my $contig_file_name = $contig_name . ".fasta";

    # my @args = ("lsrun -v clustalw -infile=" . $output_name . "/" . $file_name . " -outfile= " . $output_name . "/" . "$aln_file_name -outorder=INPUT -type=DNA");
    my @args = ("clustalw -infile=" . $output_name . "/" . $file_name . " -outfile= " . $output_name . "/" . "$aln_file_name -outorder=INPUT -type=DNA");
    system (@args);
  }

  print "alignment files done\n these files (*.aln) can be found in the $output_name directory\n";
}

# useless - can't ESTate doesn't generate the consensus sequence for each cluster - cons can do it from the clustalw files but not accurately (in particular because ESTate doesn't provide information about the strand of the cluster members)

sub run_cons {

  my ($cur_dir, $output_name, $output_name_estate) = @_;

  print "running cons on each contig to generate the consensus sequence...\n";

  my $output_name_path = $cur_dir . "/" . $output_name;
  # print "output_path: $output_name_path.\n";
  
  # output directory files
  opendir (THISDIR, $output_name_path) or die "can not read this directory $output_name_path ...";
  my @files = grep {$_ =~ /^cluster.*aln/} readdir THISDIR;

  closedir THISDIR;
  
  foreach my $aln_file_name (@files) {
    my @tmp = split (/\./, $aln_file_name);
    my $cluster_name = $tmp[0];
    # print "cl name: $cluster_name\n";
    my $cluster_number = substr ($cluster_name, 7, length ($cluster_name)) + 0;
    # print STDOUT "cluster_number: $cluster_number\n";
    my $contig_name = "CONTIG" . $cluster_number;
    my $contig_file_name = $contig_name . ".fasta";
    my $cluster_file_name = $cluster_name . ".fasta";

    my @args = ("cons $output_name" . "/" . "$aln_file_name -name $contig_name -outseq $output_name" . "/" . "$contig_file_name");
    system (@args);
    
    my $new_cluster_file_name_path = $output_name_estate . "/" . $cluster_file_name;
    my $cluster_file_name_path = $output_name . "/" . $cluster_file_name;
    my $contig_file_name_path = $output_name . "/" . $contig_file_name;
    @args = ("cat $cluster_file_name_path $contig_file_name_path > $new_cluster_file_name_path");
    system (@args);
  }
}

sub gather_contig_info {

  my ($cur_dir, $output_name) = @_;

  my @a_info;

  my $output_name_path = $cur_dir . "/" . $output_name;
  # print "output_path: $output_name_path.\n";
  
  # output directory files
  opendir (THISDIR, $output_name_path) or die "can not read this directory $output_name_path ...";
  my @files = grep {$_ =~ /^cluster.*fasta/} readdir THISDIR;
  
  # print STDOUT "files: \n@files\n";

  closedir THISDIR;
  
  foreach my $file_name (@files) {
    my $file_name_path = $output_name_path . "/" . $file_name;
    # print "file: $file_name_path\n";
    my @tmp = split (/\./, $file_name);
    my $cluster_name = $tmp[0];
    # print "cl name: $cluster_name\n";
    my $cluster_number = substr ($cluster_name, 7, length ($cluster_name)) + 0;
    # print STDOUT "cluster_number: $cluster_number\n";
    my $contig_name = "CONTIG" . $cluster_number;
 
    my $contig_info = $contig_name;
    my @members_list = ();
    my $sequence = "";
    my $getSequence = "false";

    open (INFILE, "< $file_name_path") or die "can't open this file: $!";
    while (<INFILE>) {
      chomp;
      if ($_ =~ /^>/) {
	my @a_tmp = split (/ /, $_);
	my $member_name = substr ($a_tmp[0], 1, length($a_tmp[0]));
	# print "member name: $member_name\n";
	if ($_ =~ /MEMBERS/i) {
	  $getSequence = "true";
	  # this is not a member so we don't have to add it into the members list
	}
	else {
	  # this is a member
	  push (@members_list, $member_name);
	}
      }
      elsif (lc ($getSequence) eq "true") {
	# the sequence is not stored into a string but into an array of lines.
	# each line is a part of the whole CONTIG sequence
	$sequence = $sequence . $_;
      }
    }
    close (INFILE);

    chomp ($sequence);    

    my $nb_members = @members_list + 0;
    # print "found $nb_members members for contig $contig_name\n";
    my $line_info = ">" . $contig_name . " " . $nb_members . " members";
    if ($nb_members > 0) {
      my $i = 0;
      my $member = $members_list[$i];
      $line_info = $line_info . ", constructed from $member";
      $i ++;
      while ($i < $nb_members) {
	$member = $members_list[$i];
	$line_info = $line_info . ", $member";
	$i ++;
      }
    }
    $line_info = $line_info . "\n";
    
    # build the hash element

    my %h_element = (contig_name => "$contig_name",
		     members     => [@members_list],
		     line_info   => "$line_info",
		     sequence    => "$sequence",
		    );

    # print "members list       = ", join(', ', @members_list), "\n";
    # print "h_element{members} = ", Data::Dumper::Dumper ($h_element{'members'});

    # add data

    push (@a_info, \%h_element);
  }

  return @a_info;
}


sub gather_singlets_info {

  my ($singletons_file_path, $add_singlet_tag) = @_;

  my @a_info;

  my $line_info = "";
  # my @sequence_array;
  my $sequence;
  my $new_data = "false";

  print STDERR "parsing singlets file, $singletons_file_path...\n";

  open (INFILE, "< $singletons_file_path") or die "can't open this file, $singletons_file_path: $!";
  while (<INFILE>) {
    if ($_ =~ /^>/) {
      if (lc ($new_data) eq "true") {
	# add data
	# my $aref = [@sequence_array];
	my %h_element = (line_info => "$line_info",
			 sequence => "$sequence",
			);
	push (@a_info, \%h_element);
      }

      # new data
      chomp;
      $line_info = $_;
      if ($add_singlet_tag =~ /true/i) {
        $line_info = $line_info . " SINGLET";
      }
      $line_info = $line_info . "\n";
      # @sequence_array = ();
      $ sequence = "";
      $new_data = "true";
    }
    else {
      chomp;
      # push (@sequence_array, $_);
      $sequence = $sequence . $_;
      chomp ($sequence);
    }
  }
  close INFILE;

  # my $aref = [@sequence_array];
  my %h_element = (line_info => "$line_info",
		   sequence => "$sequence",
		  );
  push (@a_info, \%h_element);
  return @a_info;
}


sub unique_fasta_file_generation {
  
  my ($cur_dir, $input_file_base, @a_info) = @_;

  my $fasta_file_name = $input_file_base;
  
  print "generation of fasta file $fasta_file_name .. - " . (@a_info+0) . " entries\n";
  
  open (OUTFILE, "> $fasta_file_name") or die "can't create this file: $!";
  foreach my $href (@a_info) {
    print OUTFILE $href->{line_info};
    # my $aref = $href->{sequence};
    # foreach my $element (@$aref) {
    # print OUTFILE $element;
    # }
    # write the sequence into the file
    # each line has the same specified length
    my $sequence = $href->{sequence};
    $sequence = uc ($sequence);
    my $lgth = length ($sequence);
    my $i = 0;
    my $size = 60;
    while ($i < $lgth) {
      print OUTFILE substr ($sequence, $i, $size);
      print OUTFILE "\n";
      $i = $i + $size;
    }
  }
  close (OUTFILE);
  
  print "generation of fasta file $fasta_file_name done\n";

  # Indexing the result fasta database for blast search
  
  print "indexing $fasta_file_name for blast2...\n";
  
  my $fasta_file_name_path = $cur_dir . "/" . $fasta_file_name;
  my @args = ("formatdb -p f -i $fasta_file_name_path");
  system (@args);
  
  print "indexing $fasta_file_name done\n";
  
}  


sub concat {
  my ($aref1, $aref2) = @_;
  my @array1 = @$aref1;
  my @array2 = @$aref2;
  
  foreach my $element (@array2) {
    push (@array1, $element);
  }
  
  return @array1;
  
}


sub is_in {
  my ($element, $aref) = @_;
  my @list = @$aref;

  foreach my $e2 (@list) {
    if ($element eq $e2) {
      return "true";
    }
  }
  # print STDERR "$element not in the list, @list\n";
  return "false";
}

# return the h_element of an array, matching the contig name

sub get_from_aof {
  my ($contig_name, @a_info) = @_;
  my $a_length = (@a_info+0);
  my $i=0;
  while ($i<$a_length) {
    my $href = $a_info[$i];
    my %h_element = %$href;
    if ($contig_name eq $h_element{contig_name}) {
      return %h_element;
    }
    $i++;
  } 
  return undef;
}
