#! /usr/local/bin/perl
#######################################################################
# This software has been created by Genome Research Limited (GRL).    # 
# GRL hereby grants permission to use, copy, modify and distribute    # 
# this software and its documentation for non-commercial purposes     # 
# without fee at the user's own risk on the basis set out below.      #
# GRL neither undertakes nor accepts any duty whether contractual or  # 
# otherwise in connection with the software, its use or the use of    # 
# any derivative, and makes no representations or warranties, express #
# or implied, concerning the software, its suitability, fitness for   #
# a particular purpose or non-infringement.                           #
# In no event shall the authors of the software or GRL be responsible # 
# or liable for any loss or damage whatsoever arising in any way      # 
# directly or indirectly out of the use of this software or its       # 
# derivatives, even if advised of the possibility of such damage.     #
# Our software can be freely distributed under the conditions set out # 
# above, and must contain this copyright notice.                      #
#######################################################################
#
# $Header: /nfs/pathsoft/repository/psu/genexec/perl/src/exp2fasta.pl,v 1.10 1999/10/22 13:48:21 adh Exp $
#
# purpose: This script generates a FASTA file from a set of experiment
# purpose: files.
#

use Exptools;
use FileHandle;

require "badgerGetOpt.pl";


############################################################
#### parse command line - switches
############################################################
$prog="exp2fasta";
$usage="Usage: $prog -chr CHROMOSOME -match PATTERN > Fasta-file\n";

$opts = {
    "-chr" => {
	"type" => "text",
	"var" => \$chr,
	"help" => "Chromsome name",
	} ,
    "-match" => {
	"type" => "text",
	"var" => \$pattern,
	"help" => "Filename pattern to match",
	} ,
    "-searchsubdirs" => {
	"type" => "boolean",
	"var" => \$searchsubdirs,
	"def" => 1,
	"help" => "Search sub-directories of the form NNNN (where N is a decimal digit)"
	},
    "-simple" => {
	"type" => "boolean",
	"var" => \$simpleheader,
	"def" => 0,
	"help" => "Write out only the sequence name in the FASTA header"
	},
    "-escapedots" => {
	"type" => "boolean",
	"var" => \$escapedots,
	"def" => 1,
	"help" => "Interpret full-stops in patterns literally"
	},
    "-anchorhead" => {
	"type" => "boolean",
	"var" => \$anchorhead,
	"def" => 1,
	"help" => "Treat pattern as the start of the filename"
	},
    "-clip" => {
	"type" => "boolean",
	"var" => \$clip,
	"def" => 1,
	"help" => "Clip sequences before writing to FASTA file"
	}
};

&badgerGetOpt($prog,$usage,$opts) && die "Quit\n";

if ($chr) {
    $homedir = &FindHome($chr);
    die "-chr argument is not recognised" unless $homedir;
    print STDERR "Home directory is $homedir\n" if $debug;
    chdir($homedir);
}

die "-match argument missing" unless $pattern;

$pattern =~ s/\./\\\./g if $escapedots;

$pattern = "^$pattern" if $anchorhead;

print STDERR "Search pattern is $pattern\n" if $debug;

if ($searchsubdirs) {
    opendir(THISDIR,".");
    @subdirs = grep(/^\d\d\d\d$/,readdir(THISDIR));
    closedir(THISDIR);

    die "No sub-directories" unless (scalar(@subdirs) > 0);

    print STDERR "Found sub-directories: ",join(',',@subdirs),"\n" if $debug;

    foreach $dir (@subdirs) {
	die "$dir is not a directory" unless -d $dir;
    }
} else {
    @subdirs = ('.');
}

$nfiles = 0;
$nbases = 0;

my $exp = new Exptools();

foreach $dir (@subdirs) {
    opendir(THISDIR,$dir);
    @files = readdir(THISDIR);
    closedir(THISDIR);

    foreach $file (@files) {
	next if $file =~ /SCF$/;
	next unless $file =~ /\.\S+/;
	next unless $file =~ /$pattern/;
	my $fh = new FileHandle("$dir/$file", 'r');
	next unless $fh;
	if ($exp->readExpFile($fh)) {
	    print STDERR "WARNING -- readExpFile failed for $file.\n";
	    $fh->close();
	    next;
	} else {
	    $fh->close();
	    $rc = &ExpToFasta($exp, $dir, $file, $simpleheader, $clip);
	    next unless ($rc > 0);
	    $nfiles += 1;
	    $nbases += $rc;
	}
    }
}

print STDERR "Wrote $nbases bp in $nfiles sequences\n";

exit(0);
    
sub ExpToFasta {
    my ($exp, $dir, $file, $simplehead, $clip, $junk) = @_;
  
    # Determine Entry name
    my $read = $exp->getEntry("ID");
    chop($read);

    $read = $file unless $read;
    $read =~ s/^.*\///;

    # The sequence
    my $DNA = $exp->getEntry("SQ");

    if ($clip) {
	my ($lclip,$rclip) = $exp->determineActiveRegion();
	$lclip -= 1;
	my $cliplen = $rclip - $lclip;
	$DNA = substr($DNA, $lclip, $cliplen);
    }

    $DNA =~ tr/-*/n-/;

    my $seqlen = length($DNA);

    print ">$file len=$seqlen";
    unless ($simplehead) {
	print " dir=$dir" if ($dir ne '.');
	print " id=$read";
    }
    print "\n";

    my $linelen = 50;

    while ($DNA) {
	print substr($DNA,0,$linelen),"\n";
	$DNA = substr($DNA,$linelen);
    }

    return $seqlen;
}

sub FindHome {
    my $chr = shift;

    return 0 unless $chr;

    $chr = uc($chr);

    my ($org, $chrnum, $uname, $homedir);

    if ($chr =~ /^(MAL|TRYP)\d+/) {
	($org, $chrnum) = $chr  =~ /^([A-Z]+)(\d+)/;
	$uname = ($org eq 'MAL') ? "malaria" : "tryp";
    } else {
	$uname = lc($chr);
    }

    my @pwdata = getpwnam($uname);
    $homedir = $pwdata[7];

    unless ($homedir) {
	print STDERR "WARNING -- failed to map $uname to home directory\n";
	return 0;
    }

    $homedir .= "/$chr";
    unless (-d $homedir) {
	print STDERR "WARNING -- $homedir is not a directory\n";
	return 0;
    }

    print STDERR "\n" if $debug;

    return $homedir;
}
