#!/usr/local/bin/perl5.6.0 -w

# $Header: /nfs/pathsoft/repository/psu/genexec/perl/src/gf2tab.pl,v 1.5 2002/01/03 11:35:38 kmr Exp $
# purpose: Generate a tab file from a genefinder output file
# input  : the genefinder output
# output : the features are directed to the standard output

# e.g. : gf2tab.pl gf.out > gf.tab

# version 2 => more information are parsed from the Genefinder output
# input  : the genefinder output and the sequence file
# output : a tab file with the CDS information
#          another tab file with the splice site information

# e.g. : gf2tab.pl gf.out seq.fasta
# output => gf.cds.tab and gf.ss.tab

=head1 NAME

gf2tab.pl 

=head1 SYNOPSIS

  gf2tab.pl [genefinder output file]  [fasta sequence]

  e.g. : gf2tab.pl gf.out seq.fasta

=head1 DESCRIPTION

  convert the genefinder output in a tab format

  @author axk

=cut

# BioPerl Modules
use Bio::SeqIO;

sub print_feature
{
  my ($name, $score, $complement, $cut_position, @exons) = @_;
  local $"= ',';

  $exon_string = "@exons";

  if (@exons > 1) {
    $exon_string = "join($exon_string)";
  } 

  if ($complement) {
    $exon_string = "complement($exon_string)";
  }

  if ($cut_position =~ /0/ || $cut_position =~ /A/) {
    $codon_start = 1;
  }
  elsif ($cut_position =~ /1/) {
    $codon_start = 3;
  }
  elsif ($cut_position =~ /2/) {
    $codon_start = 2;
  }

  # FT                   /note="$name  - score = $score - the cut position within the codon is $cut_position"

  print <<EOF;
FT   CDS             $exon_string
FT                   /note="$name  - score = $score"
FT                   /codon_start=$codon_start 
FT                   /score=$score
EOF
}


my $_gf         = "-gf";
my $_seq        = "-seq";

# Check if in the command line specifies a gf output (-gf) and a sequence (-seq)

my $lgth = @ARGV + 0;
if ($lgth < 4) {
  print "specify an gf output file and a sequence fasta file\n";
  display_help ();
  exit 1;
}
elsif ($lgth >= 4) {
  my $i = 0;
  foreach my $element (@ARGV) {
    if ($element eq $_gf) {
      $gf_input_name = $ARGV[$i+1];
    }
    elsif ($element eq $_seq) {
      $seq_input_name = $ARGV[$i+1];
    }
    # if at least one of the options is not recognized, display the help message and exit
    elsif ($element =~ /^-/) {
      print STDERR "option not recognized: $element\n";
      display_help ();
      exit 1;
    }
    $i ++;
  }
}
if ((not defined $gf_input_name) or not (defined $seq_input_name)) {
  print "specify an gf output file and a sequence fasta file\n";
  display_help ();
  exit 1;
}


my $cur_dir = qx/pwd/;
chomp ($cur_dir);

my $seq_input_name_path = $cur_dir . "/" . $seq_input_name;

my @a_tmp = split (/\.out/, $gf_input_name);
# this file stores the splice site information
$ss_file_name = $a_tmp[0] . ".ss.tab";
# this file stores the cds information
$cds_file_name = $a_tmp[0] . ".cds.tab";
# print STDERR "ss_file_name : $ss_file_name\n";

# Handle a sequence object - using BioPerl Module
# the purpose is to process the sequence to display the start codons and the splice sites

# BIOPerl
my $in  = Bio::SeqIO->new('-file' => "$seq_input_name_path",
                           '-format' => 'Fasta');
my $seq_obj = $in->next_seq();
my $lgth_seq = $seq_obj->length();
my $seq = $seq_obj->subseq (1, $lgth_seq);
$lgth_seq = length ($seq);
# print STDERR "seq length: $lgth_seq\n";

if (not defined $lgth_seq) {
  print STDERR "error while processing the sequence file\n";
  exit 1;
}

open (INPUT, "< $gf_input_name") or die "can't open input file : $gf_input_name\n";
open (FILESS, "> $ss_file_name") or die "can't open tab file : $ss_file_name\n";
open (FILECDS, "> $cds_file_name") or die "can't open tab file : $cds_file_name\n";

select FILECDS;

while (<INPUT>) {
  # print STDERR "line in while : $_\n";
  if (/^(\S+):$/ && ! /ORFs:/) {
    $hit_name = $1;
    # print STDERR "hit_name defined : $hit_name\n";
  }
  elsif (/ORF:/) {
    # ORF information in the current line
    # $orf_info = $_;
    # print STDERR "orf_info defined : $_\n";
    
    /.+:\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+\.\d+)\s+(\d+\.\d+)\s+(\d+\.\d+)/;
    # $start    = $1;
    # $stop     = $2;
    # $length   = $3;
    # $scoreSS  = $4;
    # $coding_segment_score = $5;
    # $combined_score       = $6;
    
    # print STDERR "line : $_\n";
    # print STDERR "start, stop, length, scoreSS score2 score3 : $start $stop $length $scoreSS $coding_segment_score $combined_score\n";
    
    # the following line is the ORF translation
    $new_line = <INPUT>;
    
    # the following lines are the splice site information
    # e.g. :          1201   659   543  11.90 
    $new_line = <INPUT>;
    
    # print STDERR "SS line out loop : $new_line\n";
    while (defined $new_line && $new_line =~ /.+/) {
      # $new_line = $_;
      # print STDERR "SS line in loop : $new_line\n";
      if ($new_line =~ /\s+\d+\s+\d+\s+\d+\s+\d+\.\d+/) {
	# print STDERR "this line is rubbish : $new_line\n";
      }
      
      # Splice site information

      if ($new_line =~ /(\*\s+|\s+)(0\s+|1\s+|2\s+|\s+)(3'|5'|atg)\s+(\d+)\s+(\d+\.\d+)/) {
	# the "asterix" if present
	# $asterix = $1;

	# the position base at which the site interrupts the codon
	# $base = $2;

	# the site type
	$site_type = $3;

	# the position of the splice site is the position of the nearest base of the corresponding exon
	# i.e. the first base of the exon for a 3' splice site, and the last exon base for a 5' splice site.
	$position = $4;

	# unnormalized LLR score
	$LLRscore = $5;
	
	# this information are going to another tab file
	
	# print STDERR "line in loop : $new_line\n";
	# print STDERR "asterix, base, site_type, position, LLRscore : $asterix, $base, $site_type, $position, $LLRscore\n";
	
	# writing splice site information in a tab file
	if ($site_type =~ /3'/) {
          # the location of the two potential 3' SS, display 5 bases : GGTAG

	  my $pos_shift = $position - 1;

	  if (lc (substr ($seq, $pos_shift-2, 2)) eq "ag") {
	    $location = ($position-5) . ".." . ($position-1);
	    if (lc (substr ($seq, $pos_shift+1, 2)) eq "ct") {
	      $location_bis = "complement(" . ($position+5) . ".." . ($position+1) . ")";
	    }
	  }
	  else {
	    $location = "complement(" . ($position+5) . ".." . ($position+1) . ")";
	  }

	  $colour = 5;
          $site_type = $site_type . " splice site";
	}
        elsif ($site_type =~ /5'/) {

          my $pos_shift = $position - 1;

          # the location of the two potential 5' SS, display 6 bases : GTTATA
          
	  if (lc (substr ($seq, $pos_shift+1, 2)) eq "gt") {
	    $location = ($position+1) . ".." . ($position+6);
	    if (lc (substr ($seq, $pos_shift-1, 2)) eq "ca") {
	      $location_bis = "complement(" . ($position-1) . ".." . ($position-6) . ")";
	    }
	  }
	  else {
	    $location = "complement(" . ($position-1) . ".." . ($position-6) . ")";
	  }
	 
	  $colour = 11;
	  $site_type = $site_type . " splice site";
	}
	else {
	  # start codon
	  
	  my $pos_shift = $position - 1;

	  if (lc (substr ($seq, $pos_shift, 1)) eq "a") {
	    $location = ($position) . ".." . ($position+2);
	  }
	  else {
	    $location = "complement(" . ($position) . ".." . ($position-2) . ")";
	  }

	  $site_type = $site_type . " start codon";
	  $colour = 6;
	}
	
	print FILESS "FT   UNSURE          $location\n";
	print FILESS "FT                   /note=\"$site_type - LLR score = $LLRscore\"\n";
	print FILESS "FT                   /score=$LLRscore\n";
	print FILESS "FT                   /colour=$colour\n";

	undef $location;

	if (defined ($location_bis)) {
	  print FILESS "FT   UNSURE          $location_bis\n";
	  print FILESS "FT                   /note=\"$site_type - LLR score = $LLRscore\"\n";
	  print FILESS "FT                   /score=$LLRscore\n";
	  print FILESS "FT                   /colour=$colour\n";
	  
	  undef $location_bis;
	}
      }
      if (! eof INPUT) {
	$new_line = <INPUT>;
      }
      else {
	undef $new_line;
      }
    }
    # print STDERR "loop over ...\n";
  }
  elsif (defined $hit_name) {
    if (m/\s+([\d\.]+)\s+(?:([\d\.]+)\s+)?([A012]) (.*\d\.\.\d*)/) {
      $score = $1;

      my $cut_position = $3;

      # print STDERR "arg 3 and cut_position : $3, $cut_position\n";

      # first line of an exon list
      while (/,\s+$/) {
	# read another line
	# print STDERR "line in sub_while : $_\n";
	$new_line = <INPUT>;
	if (length $new_line == 0) {
	  s/,$//;
	  } else {
	    $_ .= $new_line;
	  }
      }
      
      # we have now collected the whole line
      (@exons) = m/(\d+\.\.\d+)/g;
      
      $exons[0] =~ m/(\d+)\.\.(\d+)/;
      
      $complement = 0;
      
      if ($1 > $2) {
	$complement = 1;
	@exons = map {
	  s/(\d+)\.\.(\d+)/$2..$1/;
	  $_;
	} reverse @exons;
      }
      print_feature $hit_name, $score, $complement, $cut_position, @exons;
    }
    undef $hit_name;
  }
} # end while

close (FILESS);
close (FILECDS);

###################################################################
# The end
###################################################################

 # display the help message

sub display_help {
  print "gf2tab.pl -gf [gf_output_file] -seq [seq_file] \n\n";
  print "\t-gf the genefinder output file\n";
  print "\t-seq the sequence in a fasta file\n";
}
