#!/nfs/newlocal/bin/perl5.6.0 -w

use strict;

use Bio::PSU::Range;
use Bio::PSU::Seq;
use Bio::PSU::Feature;
use Bio::PSU::IO::EMBL::Stream;
use Bio::PSU::IO::BufferFH;
use Bio::PSU::SeqFactory;

if ((@ARGV+0) == 0) {
  print STDERR "specify the name of a hmmgene output file\n";
  exit 0;
}

my $gffFile       = $ARGV[0];
my $seqFile       = $ARGV[1];
my @tmp           = split ('\.', $gffFile);
my $cdsFile       = $tmp[0] . ".hmmgene.cds.tab";
my $otherFeatFile = $tmp[0] . ".hmmgene.rest.tab";

open HMMGENE, "<$gffFile"       or die "can't open file, $gffFile\n";
open CDS,     ">$cdsFile"       or die "can't open file, $cdsFile\n";
open OTHER,   ">$otherFeatFile" or die "can't open file, $otherFeatFile\n";

my @features = ();
my $seq      = Bio::PSU::Seq->new (-type => 'dna');

my $seqi     = Bio::PSU::SeqFactory->make(-file => $seqFile,    -format => 'fasta');
my $seq_obj  = $seqi->next_seq;

my $bfh      = Bio::PSU::IO::BufferFH->new(-fh => \*CDS);
my $stream   = Bio::PSU::IO::EMBL::Stream->new(-bfh => $bfh);
# start with, ie with a UTR => complete gene or not => the first may be missing and we have to work out the frame
my $start_with = undef;

while (<HMMGENE>) {
  next if /^#|SEQUENCE/;
  my $line = $_;
  # print STDERR "new line - expecting a UTR : $line";
  $line =~ /(\w+)\t+(\S+)\t+(\w+)\t+(\d+)\t+(\d+)\t+(\S+)\t+(-|\+)\t+(\S+)/;
  my $seqName     = $1;
  my $nameFeature = $3;
  my $start       = $4;
  my $end         = $5;
  my $score       = $6;
  my $strand      = $7;
  my $frame       = $8;

  $seq->id ($seqName);

  # CDS ranges
  my @ranges = ();
  # exon scores
  my @scores = ();
  my $total_scores = 0;
  my $nb_exons = 0;

  # if one or several UTRs, copy them into the tab file

  if (($nameFeature =~ /UTR/)) {
    # the UTR
    print OTHER "$line";
    if (not defined ($start_with)) {
      $start_with = "UTR";
    }
    next;
  }
  # build the cds feature by joining the exons together
  # bear it can be an unfinished sequence and so the UTR can be missing
  # it that case, will stop on the SEQUENCE FEATURE
  
  if (not defined ($start_with)) {
    $start_with = "coding";
  }

  while (not eof (HMMGENE)) {
    # print STDERR "sub-loop on line: $line\n";
    if ($line =~ /^#|SEQUENCE/) {
	$line = <HMMGENE>;
	next;
      }
    if ($line =~ /UTR/) {
      last;
    }
      $line =~ /(\w+)\t+(\S+)\t+(\w+)\t+(\d+)\t+(\d+)\t+(\S+)\t+(-|\+)\t+(\S+)/;
      my $seqName     = $1;
      my $nameFeature = $3;
      my $start       = $4;
      my $end         = $5;
      my $score       = $6;
      my $strand      = 1;
      if ($7 =~ /-/) {
	$strand = -1;
      }
      my $frame       = $8;
      
      if ($nameFeature =~ /coding/i) {
	# exon
	my $range = Bio::PSU::Range->new(-start       => $start,
					 -end         => $end,
					 -strand      => $strand);
	push (@ranges, $range);
	push (@scores, $score);
	$nb_exons++;
	$total_scores = $total_scores + $score;
      }
      else {
	# intron
	print OTHER "$line";
      }
      $line = <HMMGENE>;
    } # end while
      
    # the second UTR
    if ($line =~ /UTR/) {
      print OTHER "$line";
    }
      if ($nb_exons > 0) {
	my $cds_score = $total_scores / $nb_exons;
	$cds_score = round ($cds_score, 2);
	
	# build the feature object
	
	my $feature = Bio::PSU::Feature->new(-key    => 'CDS',
					     -ranges => \@ranges);
	my $note = "HMMGene - exon scores: " . join (', ', @scores);

	if ($start_with eq "coding") {
	  # work out the frame - input the sequence of the first predicted exon
	  print STDERR "working out the frame ...\n";
	  $frame = getFrame ($ranges[0], $seq_obj);
	}
	
	# specify the frame for the first exon, which might not start with a MET
	my $codon_start = undef;
	if (defined ($frame)) {
	  if ($frame =~ /0/) {
	    $codon_start = 1;
	  }
	  elsif ($frame =~ /1/) {
	    $codon_start = 2;
	  }
	  elsif ($frame =~ /2/) {
	    $codon_start = 3;
	  }
	}

	$feature->qadd('note', $note);
	$feature->qadd('method', 'HMMGENE');
	$feature->qadd('score', $cds_score);
	$feature->qadd('colour', '6');
	if (defined $codon_start) {
	  $feature->qadd('codon_start', $codon_start);
	}
	
	push (@features, $feature);
      }
} # end while

$seq->features (@features);
$stream->write_seq($seq);

#######################################################################

sub round {
  my ($number, $nb_decimal) = @_;
  my $length_string = length ($number);
  if ($number =~ /\./) {
    $number =~ m/\./g;
    my $length = pos ($number);
    $length = $length + $nb_decimal;
    my $new_number = substr ($number, 0, $length);
    # print STDERR "number, new_number: $number, $new_number\n";
    return $new_number;
  }
  else {
    return $number;
  }
}

sub getFrame {
  my ($range, $seq_obj) = @_;
  my $frame = undef;
  
  my $subseq = getSubSequence ($seq_obj, $range);
  my $frame_index = 0;
  my $seq_length = length ($subseq);
  my $found_frame = 0;

  while (not ($found_frame) && ($frame_index < 3)) {
    print STDERR "working on the frame $frame_index\n";
    my $i = $frame_index;
    my $is_stop_codon = 0;
    while (($i<$seq_length-2) && not $is_stop_codon) {
      my $codon = substr ($subseq, $i, 3);
      if ($codon =~ /taa|tag|tga/i) {
	print STDERR "found a stop codon on frame $frame_index, $codon\n";
	$is_stop_codon = 1;
      }
      $i = $i + 3;
    }
    if (not $is_stop_codon) {
      $found_frame = 1;
      $frame = $frame_index;
    }
    $frame_index++;
  }

  print STDERR "frame found (yes/no) ? $found_frame\n";

  return $frame;
}

sub getSubSequence {
  my ($seq_obj, $range) = @_;

  my $sub_seq = $seq_obj->subseq ($range->start, $range->end, $range->strand);
  return $sub_seq->str;
}
