#!/usr/local/bin/perl5.6.0 -w

=head1 NAME

ipro2tab - Splice-friendly annotator of InterPro search results onto
EMBL feature tables

=head1 SYNOPSIS

 ipro2tab <embl file> [-s int] <InterPro raw file>

=head1 DESCRIPTION

This program takes the raw file from an InterPro search run (named by
InterProScan as merged.raw, by default) and annotates the hits onto a
corresponding EMBL feature table. The mapping of result to CDS is made
by gene name. Individual search hits are annotated as 'misc_feature'
while the InterPro accession numbers are added to the CDS feature
itself as 'db_xref'. In effect this makes a modified copy of the
original feature table. The output is an EMBL feature table sent to
STDOUT, plus a feature table for each database, containing the hit
annotation.

The -s switch changes the style of annotation produced.

Note that you may see warnings about features being clipped to fit
into a CDS. This is because of bugs in InterProScan where it reports
hits outside of the CDS range. Any hit features clipped like this will
have a note saying how many amino acids were clipped from the end.

The hit annotations do not contain any scores; this is an omission
from InterProScan which appears to be a design decision.

=head1 METHODS

None

=head1 AUTHOR

Keith James (kdj@sanger.ac.uk)

=head1 COPYRIGHT

Copyright (C) 2001, 2002 Genome Research Limited. All Rights Reserved.

=head1 DISCLAIMER

This software is provided "as is" without warranty of any kind. It may
be used, redistributed and/or modified under the same conditions as
Perl itself.

=cut

use strict;
use Getopt::Std;
use Bio::PSU::SeqFactory;
use Bio::PSU::FeatureFactory;

my $ipr_home = "/nfs/pathsoft/external/bio-soft/iprscan";
my $profile  = "$ipr_home/data/prosite_prerelease.prf";
my $smart    = "$ipr_home/data/smart.HMMs";
my $tigrfam  = "$ipr_home/data/TIGRFAMs_1.2_HMM.LIB";

my $usage = "Usage: ipro2tab  [-s 0|1] <embl file> <InterPro raw file>

\t-s Set annotation style  (default 0)

Annotation styles are:

\t0) Regular style
\t1) Eukaryote style\n";

my (%opts, $style, $embl_file, $ipr_file);
getopts('hs:', \%opts);

defined $opts{h} and die $usage;
defined $opts{s} and $style = $opts{s};

unless (defined $style) { $style = 0 }
unless ($style == 0 || $style ==1) { die $usage}
unless (@ARGV == 2) { die $usage }

$embl_file = shift;
$ipr_file  = shift;

print STDERR "Reading EMBL entry $embl_file\n";
my $es = Bio::PSU::SeqFactory->make(-file => "<$embl_file");
my $e  = $es->next_seq;

my $ff = Bio::PSU::FeatureFactory->new;

# Read in the accessions/descriptions of the Prosite pre-release
# patterns as they can't be obtained from SRS

my %profiles;

open(PRO, "<$profile") or die "Unable to open $profile for reading: $!\n";
print STDERR "Fetching descriptions of Prosite pre-release profiles\n";

my $pac;

while (<PRO>)
{
    chomp;
    if (/^AC\s+(\S+);/)
    {
		$pac = $1
    }
    if (/^DE\s+(.*)/)
    {
		$profiles{$pac} = $1;
        print STDERR ".";
    }
}

print STDERR "\n";
close(PRO);

# Read in the accessions/descriptions of the Smart HMMs as they can't
# be obtained from SRS

my %smart;

open(SMA, "<$smart") or die "Unable to open $smart for reading: $!\n";
print STDERR "Fetching descriptions of Smart HMMs\n";

my $sac;
while (<SMA>)
{
    chomp;
    if (/^ACC\s+(\S+)/)
    {
		$sac = $1
    }
    if (/^DESC\s+(.*)/)
    {
		$smart{$sac} = $1;
		print STDERR ".";
    }
}

print STDERR "\n";
close(SMA);

# Read in the accessions/descriptions of the TIGR HMMs as they can't
# be obtained from SRS

my %tigrfam;

open(TGR, "<$tigrfam") or die "Unable to open $tigrfam for reading: $!\n";
print STDERR "Fetching descriptions of TIGRFAM HMMs\n";

my $tac;
while (<TGR>)
{
    chomp;
    if (/^ACC\s+(\S+)/)
    {
		$tac = $1
    }
    if (/^DESC\s+(.*)/)
    {
		$tigrfam{$tac} = $1;
		print STDERR ".";
    }
}

print STDERR "\n";
close(TGR);

# Read in the raw InterPro search output
open(IPR, "<$ipr_file") or die "Unable to open $ipr_file for reading: $!\n";
print STDERR "Reading InterPro hits:\n";

my %interpro;

while (my $line = <IPR>)
{
    # File is tab-delimted:
    #
    # $cds:      CDS/gene name (from Fasta header)
    # $checksum: Checksum
    # $len:      CDS length in amino acids
    # $program:  Search program
    # $proacc:   Accession number in seach program's native database
    # $short:    Short name of motif
    # $start:    Motif start position in CDS
    # $end:      Motif end position in CDS
    # $score:    Search program score
    # $psconf:   For Prosite - was match verified by a Confirm pattern?
    # $date:     Date search executed
    # $ipracc:   InterPro accession number
    # $desc:     InterPro entry description

    my ($cds, $checksum, $len, $program, $proacc, $short, $start, $end, $score, $psconf, $date, $ipracc, $desc)
		= split(/\t/, $line);

    # There may be multiple hits per cds
    $interpro{$cds} = [] unless exists $interpro{$cds};

    push(@{$interpro{$cds}},
		 [$checksum, $len, $program, $proacc, $short, $start, $end, $score, $psconf, $date, $ipracc, $desc]);
    print STDERR ".";
}

print STDERR "\n";
close(IPR);

# New annotation will be stored here as Bio::PSU::Features
my %iprfeatures = (BlastProDom => [],
				   FPrintScan  => [],
				   HMMPfam     => [],
				   ProfileScan => [],
				   ScanRegExp  => [],
				   HMMSmart    => [],
                   HMMTigr     => []);

# Define the colours of the new features
my %iprcolours = (BlastProDom => 0,
				  FPrintScan  => 4,
				  HMMPfam     => 9,
				  ProfileScan => 7,
				  ScanRegExp  => 8,
				  HMMSmart    => 13,
                  HMMTigr     => 5);

# Define the real names for the searches
my %iprnames = (BlastProDom => "prodom",
				FPrintScan  => "prints",
				HMMPfam     => "pfam",
				ProfileScan => "prosite pattern",
				ScanRegExp  => "prosite profile",
				HMMSmart    => "smart",
                HMMTigr     => "tigrfam");

# Define the SRS databases for description lookup
my %known_SRS = (FPrintScan  => [qw(prints  des)],
				 HMMPfam     => [qw(pfam    des)],
				 ScanRegExp  => [qw(prosite des)]);

# Foreach feature in the searched EMBL entry
print STDERR "Annotating: + = processing feature, * = getz lookup:\n";
foreach my $f ($e->features)
{
	next unless $f->key eq 'CDS';

	my $data;
    foreach ($f->gene)
    {
		$data = delete $interpro{$_};
		last if defined $data;
    }
    next unless defined $data;

    my %iprhits;
    my @domains;

    foreach my $result (@$data)
    {
		my ($checksum, $len, $program, $proacc, $short, $start, $end, $score, $psconf, $date, $ipracc, $desc) = @$result;

		# Store non-NULL InterPro accession numbers associated with
		# hit(s) with their description
		if (defined $ipracc and $ipracc ne 'NULL')
		{
			$iprhits{$ipracc} = $desc
		}

		# Recover from a bad feature and continue after printing
		# backtrace
		my $iprf = undef;

		eval { $iprf = $ff->annotate_translation($f,
												 $start,
												 $end - $start + 1) };

		if ($@)
		{
			warn "Recieved error from FeatureFactory: $@\n"
		}

		print STDERR "+";
		next unless defined $iprf;

		# Hash lookup to find description of pre-release Prosite
		# profile/Smart or SRS lookup, if available
		my $family;

	SWITCH:
		{
			if ($program eq 'ProfileScan')
			{
				$family = $profiles{$proacc};
				last SWITCH;
			}

			if ($program eq 'HMMSmart')
			{
				$family = $smart{$proacc};
				last SWITCH;
			}

            if ($program eq 'HMMTigr')
			{
				$family = $tigrfam{$proacc};
				last SWITCH;
			}

			if ($program eq 'ScanRegExp')
			{
				$family = get_family_desc(\%known_SRS, $program, $proacc);

				print STDERR "*";

				if ($psconf eq 'T')
				{
					$family .= " Confirmed by InterPro eMOTIF pattern match."
				}
				last SWITCH;
			}

            $family = get_family_desc(\%known_SRS, $program, $proacc);
			print STDERR "*";
		}

        if (! defined $family)
        {
            $family = "<Unable to resolve family name>";
        }

		if ($style == 0)
		{
			$iprf->qadd('note', "$program hit to $proacc, $family, score $score");
			$iprf->qadd('colour', $iprcolours{$program});
			$iprf->qadd('label', $program);
		}
		elsif ($style == 1)
		{
            my $dom = $iprnames{$program} . ":$proacc;$family;$score;codon $start-$end";
            push(@domains, $dom);

			$iprf->qadd('psu_domain', $dom);
			$iprf->qadd('colour', $iprcolours{$program});
			$iprf->qadd('label', $program);
		}

		push(@{$iprfeatures{$program}}, $iprf);
    }

    # Add a xref to each feature in the searched EMBL entry which has
    # been hit
    foreach my $ipracc (keys %iprhits)
    {
		if ($style == 0)
		{
			$f->qadd('db_xref', $ipracc)
		}
		elsif ($style == 1)
		{
			$f->qadd('psu_db_xref', "interpro:$ipracc;" . $iprhits{$ipracc});

            foreach (@domains)
            {
                $f->qadd('domain', $_);
            }
		}
    }
}
print STDERR "\n";

foreach my $database (keys %iprfeatures)
{
    my $fo = Bio::PSU::SeqFactory->make(-file   => ">interpro_$database.tab",
										-format => 'embl');

    print STDERR "Writing interpro_$database.tab\n";
    # As the hash already holds a reference to a list of features, we
    # don't need to dereference it
    $fo->write_seq(Bio::PSU::Seq->new(-features => $iprfeatures{$database}));
}

print STDERR "Writing EMBL entry\n";
my $eo = Bio::PSU::SeqFactory->make(-fh     => \*STDOUT,
									-format => 'embl');

$eo->write_seq($e);


sub get_family_desc
{
    my ($known_db, $db, $acc) = @_;

    return $acc unless exists $known_db->{$db};

    my $srs_db = $known_db->{$db}->[0];
    my $field  = $known_db->{$db}->[1];

    open(GETZ, "getz '[$srs_db-acc:$acc]' -f $field |") or
		die "Unable to open pipe from getz";

    my $desc = "";

    while (<GETZ>)
    {
		chomp;
		s/^\S+\s+(.*)/$1/;
		$desc .= $_;
    }
    close(GETZ);

    # Strip leading/trailing whitespace
    $desc =~ s/^\s+//;
    $desc =~ s/\s+$//;

    # Strip terminal . or ;
    $desc =~ s/[.;]$//;

    return $desc;
}
