#!/usr/local/bin/perl -w

# supply path for additional perl modules
BEGIN{ push(@INC, "/nfs/disk222/yeastpub/bio-soft/lib/site_perl/5.6.0") }
#BEGIN{ push(@INC, "/nfs/disk222/yeastpub/scripts/perl/bioperl-06") }


# supply paths for blast database and newer version of primer3
#my $BLAST_DIR = "/nfs/disk222/yeastpub/people/MIKE/SCRIPTS/PRIMER_SCRIPTS/POMBE_CDS/RESULTS/CONTIGS";
my $BLAST_DIR = "/nfs/disk222/yeastpub/blast_server/Sequence_Databases/";
my $primer3 = "/nfs/disk222/yeastpub/people/MIKE/PRIMER3/primer3_0_9_test/src/primer3_core";

# Specify modules
use strict;
use Getopt::Std;
use Bio::SeqIO;
use Bio::Seq;
use Bio::Tools::Blast;
use Bio::PSU::Seq;
use Bio::PSU::SeqFactory;

my @entries;
my @bio_entries;
my (%opts, $format, $verbose);

my $dup = ".dup";
my $null_gene_string = "undefined_gene";
my $null_prod_string = "undefined product";

my $usage = "Usage: primer_picker.pl [-v] [-m <format>] file_of_filenames

\t-m write format (Fasta, EMBL, GCG or Raw)
\t-v print verbose warnings
";

###
### Command line options
###

getopts('hm:v', \%opts);

defined $opts{"h"} and die $usage;
defined $opts{"v"} and $verbose++;

defined $opts{"m"} and $format = $opts{"m"};
if (defined $format)
{
    unless ($format =~ /raw|fasta|embl|gcg/i)
    {
	die "Invalid output format: $format\n"
    }
}
else
{
    print STDERR "No output format specified. Defaulting to Fasta\n";
    $format = 'fasta';
}
my @files =();
my $contig_files = $ARGV[0];
open (CONTIG_FILES, "< $contig_files") || die "cannot open $contig_files: $!\n";
chomp(@files = <CONTIG_FILES>);
close (CONTIG_FILES);

#my @files = @ARGV; # -for supplying files on command line
unless (@files) {die $usage};

###
### Process the EMBL files
###

foreach my $file (@files)
{
    print STDERR "Reading $file...\n" if $verbose;
;
    my @new_entries = read_db($file, 'embl');
    push(@entries, @new_entries);
}

# Output files: progress log, errors, processed primers
open (LOG_FILE, "> pombe_process_log") || die "cannot open pombe_process_log: $!\n";
open (FAILED_FILE, "> FAILED_genes") || die "cannot open FAILED_genes: $!\n";
open (PRIMER_FILE, "> primer_orders") || die "cannot open primer_orders: $!\n";

###
### Process the EMBL entries
###

 ENTRY_LOOP: foreach my $entry (@entries)

{
  
    my @features = $entry->features();
    

  FEATURE_LOOP: foreach my $feature (@features) 
    {
        # Look for CDS features
        if ($feature->key() =~ /CDS/) # Look for CDS features
	{
	    my @exon = get_exon_feat($feature);
	    my ($gene_string, $prod_string, $parent) = cds_details($feature, $entry);
	    my $count = 1;
	    my ($too_short, $processed) = 0;
	    my $failed_blast_count = 0;

	    print LOG_FILE "\nProcessing $gene_string (in parent:$parent)\n";

	    # Features with a /pseudo qualifier
	    if (check_cds($feature, 'pseudo')) 
	    {
		print STDERR "WARNING:$gene_string (in parent:$parent) contains a /pseudo qualifier\n" if $verbose;
		print FAILED_FILE "GENE:$gene_string\tFAILED\t/pseudo qualifier\t (in parent:$parent)\n";
		print LOG_FILE "GENE:$gene_string (in parent:$parent) - /pseudo qualifier\n";
		next FEATURE_LOOP;
	    }

	    # Features with a /partial qualifier or with < and/or > in their location
	    if (check_cds($feature, 'partial') || check_cds($feature, '_part_feature')) 
	    {
		print STDERR "WARNING:$gene_string (in parent:$parent) is only a partial sequence\n" if $verbose;
		print FAILED_FILE "GENE:$gene_string\tFAILED\t partial sequence\t (in parent:$parent)\n";
		print LOG_FILE "GENE:$gene_string (in parent:$parent) - partial sequence\n";
		next FEATURE_LOOP;
	    }

	    ###
	    ### Process the Exons
	    ###

	  EXON_LOOP: foreach my $exon (@exon) 
	    {
		my $unique_exon_id = "$gene_string" ."_" . "$count";
		my ($start, $end, $length, $exon_string, $strand) = exon_data($feature, $exon);
		print "EXON $exon_string \n";
		my $exon_number = scalar(@exon);

		# Reject exons < 400 bp
#  		if ($length < 400) 
		if ($length < 300) # relaxed cutoff
		{
		    print LOG_FILE "$unique_exon_id", "/", 
		    "$exon_number ($length bp) rejected: too short\n";
		    $too_short++;
  		} 
		else 
		{
		    print LOG_FILE "$unique_exon_id", "/", 
		    "$exon_number ($length bp): passed to primer program\n";
		}

		# Report rejections to outfiles
  		if ($too_short == $exon_number) 
		{
		    print FAILED_FILE "GENE:$gene_string\tFAILED\t All $exon_number exon(s) too short\t (in parent:$parent)\n";
  		    print LOG_FILE "No exons long enough for primer design. Process manually\n";
		    next FEATURE_LOOP;
  		} 

		# These are the exons we want
#		if ($length >= 400) 
		if ($length >= 300) # relaxed cutoff
		{

		    # Generate the boulder I/O file for each exon (inc. parameters)


		    


		    file_to_primer($unique_exon_id, $exon_string);

		    system("$primer3 < primer.temp > results.temp");


		    my ($primer_gene_id, $primer_quality, $forward_prim, $reverse_prim,
			$for_coord, $for_len, $rev_coord, $rev_len, $for_tm, $rev_tm, 
			$for_gc, $rev_gc, $prod_length);
		    my $primer_errors;
		    my ($new_for_coord, $new_rev_coord);
		    my $pcr_product;

		    # Returns an array of array references (data for each primer pair)
		    my @all_primer_data = primer_results();

		    # Loop over the primer results
		    print "PRIMERS_DONE: @all_primer_data \n";
		  PRIMER_PAIRS: for my $pri_pair_count (0 .. $#all_primer_data) 
		    {

			# Dereference, and assign primer results to an array
			my @prim_out_results = @{$all_primer_data[$pri_pair_count]};

#			print join(", ", @prim_out_results), "\n";
			
			print "PRIMER_OUT_RESULTS : @prim_out_results \n";
			
			if (@prim_out_results > 2) # i.e. exon processed successfully
			{
			    
			    # Get all of the primer data
			    ($primer_gene_id, $primer_quality, $forward_prim, $reverse_prim,
			     $for_coord, $for_len, $rev_coord, $rev_len, $for_tm, $rev_tm, 
			     $for_gc, $rev_gc, $prod_length)
				= @prim_out_results;
			    
			    print "FORWARD_PRIM: $forward_prim \n";

			    # Blast the primers to check for unique
			    write_tmp_fasta($forward_prim, $unique_exon_id);
			    system("wublastn $BLAST_DIR/contig.fasta" .
				   " temp.dna E=10000 W=10 V=3 B=1 -gap > primer_blast.out");
			    write_tmp_fasta($reverse_prim, $unique_exon_id);
			    system("wublastn $BLAST_DIR/contig.fasta" .
				   " temp.dna E=10000 W=10 V=3 B=1 -gap > primer_blast.out");

			    # Parse the Blast results
			    # Returns 1 for fail, 0 for pass
			    my $primerF_pass = prim_blast_parse($unique_exon_id, $for_len, $parent);
			    my $primerR_pass = prim_blast_parse($unique_exon_id, $rev_len, $parent);

			    # Did any of the primers pass?
			    if (($primerF_pass || $primerR_pass) 
				&& ($count == $exon_number) && ($pri_pair_count == $#all_primer_data )) 
			    {
				print LOG_FILE "GENE:$unique_exon_id\tAll primer pairs failed: Process manually\n";
				print FAILED_FILE "GENE:$gene_string\tFAILED\t PRIMER_BLAST:All primers failed\t (in parent:$parent)\n";
				last EXON_LOOP;
			    }

			    # If any of the primers failed, move on to 
			    # the next pair. Otherwise, extract the product
			    # and blast it 
			    if ($primerF_pass || $primerR_pass)
			    {
				next PRIMER_PAIRS;
			    } 
			    else 
			    {

				($new_for_coord, $new_rev_coord) = 
				    primer2parent_coord($strand, $start, $end, 
							$for_coord, $rev_coord, $prod_length);

				$pcr_product = extract_pcr_products($entry, $new_for_coord, $prod_length, $strand);
				print "PCR_PRODUCT: $pcr_product \n";
				print "FOR_COORD: $for_coord; REV_COORD: $rev_coord; NEW_FOR: $new_for_coord: NEW_REV: $new_rev_coord \n";
				print "PROD_LENGTH: $prod_length \n";
				print "START: $start END: $end \n";
				write_tmp_fasta($pcr_product, $unique_exon_id);

				system("RepeatMasker -int" . 
				       " temp.dna > temp.repeatmasker.out"); 
				
				$/ = "";
				open (MASKED_FILE, "<temp.dna.masked");
				
				MASK_LOOP: while (<MASKED_FILE>) {
				  my $line = $_;
				  print "MASKED_FILE $line \n";
				  if ($line =~ /N/) {
				    print FAILED_FILE "GENE:$unique_exon_id\tAMPLICON CONTAINS LOW COMPLEXITY SEQUENCE\t$line \n";
				    print LOG_FILE "GENE:$unique_exon_id\tAMPLICON CONTAINS LOW COMPLEXITY SEQUENCE\t$line \n";
				    $/ = "\n";
				    next PRIMER_PAIRS;
				  }
				  else {
				    $/ = "\n";
				    print "PROCESSING...\n";
				    last;
				  }
				}
				
				system("wublastn $BLAST_DIR/contig.fasta" .
				       " temp.dna V=3 B=3 > temp_blast.out");

				# Parse the Blast results
				# This time returns 1 for pass
				#           returns 0 for fail
				my $pass = blast_parse($unique_exon_id, $prod_length, 
						       $pri_pair_count, $#all_primer_data, $parent);

				if ($pass) 
				{
				    print LOG_FILE "$unique_exon_id primers processed successfully\n";

				    my $space = " "; # used with formatted output
				    
				    # Formatted output
#				    printf PRIMER_FILE "%-40.40s %-25.25s %-3.d %-6.2f %-8.4f %-3.3d %-6.d %-6.d\n", 
#				    $primer_gene_id."_F", $forward_prim, $for_len, $for_tm,
#				    $primer_quality, $prod_length, $new_for_coord, $new_rev_coord;
#				    printf PRIMER_FILE "%-40.40s %-25.25s %-3.d %-6.2f %-8.4s %-3.3s %-6.s %-6.s\n",
#				    $primer_gene_id."_R", $reverse_prim, $rev_len, $rev_tm,
#				    $space, $space, $space, $space;

				    # Tab separated output - same as from Aengus'
				    # script
				    print PRIMER_FILE $primer_gene_id."_F", 
				    "\t", lc($forward_prim), "\t$for_tm\t$for_gc\t$for_len\t$prod_length\n";
				    print PRIMER_FILE $primer_gene_id."_R",
				    "\t", lc($reverse_prim), "\t$rev_tm\t$rev_gc\t$rev_len\t$prod_length\n";
				    last EXON_LOOP;
				}
				elsif (($pass == 0) && ($count == $exon_number) && ($pri_pair_count == $#all_primer_data ))
				{
				    print LOG_FILE "GENE:$unique_exon_id\tAll amplicons failed: Process manually\n";
				    print FAILED_FILE "GENE:$gene_string\tFAILED\t AMPLICON_BLAST:".
					"All amplicons failed\t (in parent:$parent)\n";
				}
			    }
			} 
			else         # If primer script returns < 2 elements,
			{            # must be an error, quit

			    $primer_errors = @prim_out_results;
			    print FAILED_FILE "GENE:$unique_exon_id\tFAILED\tNo primers\! Error type:$primer_errors occured\n";
			    print LOG_FILE "GENE:$unique_exon_id\tFAILED\tNo primers\! Error type:$primer_errors occured\n";
			    die "GENE:$unique_exon_id\tFAILED\tNo primers\! Error type:$primer_errors occured\n";
			}

		    }

		}

		$count++;            # Increments the exon count
	    }

	    next FEATURE_LOOP;	    
	  }
      }
      
      
  }
close (LOG_FILE);
close(PRIMER_FILE);
close(FAILED_FILE);


###--------------------------------------------------------------------

###
### Subroutines
###

=head2 read_db

 Title   : read_db
 Usage   : read_db(filename, format)
 Function: opens a new SeqIO object for reading
 Example : read_db($file, 'embl')
 Returns : list of Bioperl sequence objects
 Args    : filename string, format string

=cut

sub read_db_bio
{
    my ($db_file, $db_format) = @_;
    my @db_list;

    my $db_in = Bio::SeqIO->new(-file => $db_file, '-format' => $db_format);

    while (my $bio_seq = $db_in->next_seq()) 
    {
	push(@db_list, $bio_seq);
    }
    return @db_list;
}



sub read_db
  {
   my ($db_file, $db_format) = @_;
   my @db_list;
   my $seqi = Bio::PSU::SeqFactory->make(-file => $db_file, -format => $db_format);
   #my $seqo = Bio::PSU::SeqFactory->make(-fh   => \*STDOUT, -format => 'embl');

   while (my $seq = $seqi->next_seq()) 
     {
       push(@db_list, $seq);
     }
   return @db_list;
 }




#  =head2 write_db

#   Title   : write_db
#   Usage   : write_db(filename, format, list reference)
#   Function: Opens a new SeqIO object for writing
#   Example : write_db($printfull . $dup, $format, \@dup_full)
#   Returns : Nothing
#   Args    : filename string, format string, ref to list of sequence objects

#  =cut

#  sub write_db
#  {
#      my ($db_file, $db_format, $db) = @_;

#      my $db_out = Bio::SeqIO->new(-file => ">$db_file", '-format' => $db_format);

#      foreach (@$db) 
#      {
#          $db_out->write_seq($_);
#      }
#  }


=head2 check_cds

 Title   : check_cds
 Usage   : check_cds(feature, qualifier)
 Function: Checks a feature for a qualifier (tag)
 Example : check_cds($feature, 'pseudo')
 Returns : True if present
 Args    : feature object, qualifier string

=cut

#sub check_cds
#{
#    my ($cds, $type) = @_;

#    my @tags = $cds->all_tags();
#    if (grep /$type/, @tags) 
#    {
#        return 1
#    }
#}

sub check_cds
{
    my ($cds, $type) = @_;
    my @qualifiers = $cds->qnames();
    foreach my $i (@qualifiers) {
      print "QUALIFIER: $i \n";
    }
    if (grep /$type/, @qualifiers) {
      return 1;
    }
  }




=head2 cds_details

 Title   : cds_details
 Usage   : cds_details(feature)
 Function: gets a gene name string and product string 
 Function: from a feature object
 Example : cds_details($feature)
 Returns : List of strings
 Args    : feature object

=cut

#sub cds_details
#{
#    my ($cds) = @_;
#    my (@gene, @product);
#    my $null_gene_string = "undefined_gene";
#    my $null_prod_string = "undefined product";

#    my @tags = $cds->all_tags();
#    my $parent = $cds->seq()->display_id();

#    foreach my $tag (@tags) 
#    {
#	$tag eq 'product' and @product = $cds->each_tag_value($tag);
#	$tag eq 'gene'    and @gene    = $cds->each_tag_value($tag);
#    }
#    my $gene_string = join(',', @gene);
#    my $prod_string = join(',', @product) . " (parent:" . $parent . ")";
#    $gene_string ||= $null_gene_string;
#    $prod_string ||= $null_prod_string;

#    return ($gene_string, $prod_string, $parent);
#}


sub cds_details
{
    my ($cds, $entry) = @_;
    my (@gene, @product);
    my $null_gene_string = "undefined_gene";
    my $null_prod_string = "undefined product";
    my $parent = $entry->id();
    my @gene_string  = $cds->qvalues('gene');
    my @prod_string  = $cds->qvalues('protein_id');
    my $gene_string;
    my $prod_string;
       foreach my $gene (@gene_string) {
         if ($gene =~ /\w+/) {
           $gene_string = $gene;
          }
}
       foreach my $product (@prod_string) {
         if ($product =~ /\w+/) {
           $prod_string = $product;
          }
}
	    
    $gene_string ||= $null_gene_string;
    $prod_string ||= $null_prod_string;

    return ($gene_string, $prod_string, $parent);
}







=head2 get_exon_feat

 Title   : get_exon_feat
 Usage   : get_exon_feat(feature)
 Function: retrieves sub sequence features and returns 
 Function: the exons in order
 Example : get_exon_feat($feature)
 Returns : list of strings
 Args    : feature object

=cut

#sub get_exon_feat
#{
#    my ($cds) = @_;
#    my @exon_feat = ();
#    my $strand = "";

#    if (my @exons = $cds->sub_SeqFeature()) 
#    {
#	foreach my $exon (@exons) 
#	{
#	    if ($exon->strand == -1) 
#	    {
#		$strand = "complement";
#	    }
#	    push(@exon_feat, $exon);
#	}

#    } else {
#	push(@exon_feat, $cds);
#    }

#    if ($strand eq "complement") 
#    {
#	@exon_feat = reverse(@exon_feat);
#    }

#    return @exon_feat;
#}

sub get_exon_feat
{
    my ($cds_feature) = @_;
    my @exon_feat = ();
    my $strand = "";

     $strand = $cds_feature->strand;
    if ($strand == -1){
       $strand = "complement";
     }

    if ($cds_feature->has_ranges) {
       my @ranges  = $cds_feature->ranges();
         foreach my $range (@ranges) {
         # my $str .= substr($cds_feature->{str}{val}, ($range->start - 1), ($range->end - $range->start + 1));
          push (@exon_feat, $range);
}
}
    else {
       push (@exon_feat, $cds_feature);
}

    #if ($strand eq "complement") 
    #{
#	@exon_feat = reverse(@exon_feat);
    #}

    return @exon_feat;
    


}


=head2 exon_data

 Title   : exon_data
 Usage   : exon_data(exon)
 Function: retrieves relevant data associated with exon 
 Example : exon_data($exon)
 Returns : exon data strings
 Args    : feature object

=cut

#sub exon_data
#{
#
#    my ($data) = @_;
#    my $strand = $data->strand();
#    my $start = $data->start();
#    my $end = $data->end();
#    my $length = $data->length();
#    my $exon_string = $data->seq()->seq();

#    return ($start, $end, $length, $exon_string, $strand);
#}


sub exon_data
{

    my ($feature, $exon) = @_;
    my $strand = $feature->strand();    
    #my @ranges  = $feature->ranges();
    #foreach my $range (@ranges) {
    my $start = $exon->start();
    my $end = $exon->end();
    my $length = $end - $start;
    my $exon_string .= substr($feature->{str}{val}, ($exon->start - 1), ($exon->end - $exon->start + 1));
    if ($strand == -1) {
      $exon_string = reverse($exon_string);
      $exon_string =~ tr/[GTACgtac]/[CATGcatg]/;
}
    return ($start, $end, $length, $exon_string, $strand);
}








=head2 file_to_primer

 Title   : file_to_primer
 Usage   : file_to_primer(gene id, sequence)
 Function: makes a Boulder I/O file to pass to primer3
 Example : file_to_primer($unique_gene_id, $exon_string)
 Returns : - writes a temp primer file
 Args    : gene and sequence strings

=cut

sub file_to_primer
{

    my ($unique_gene_id, $exon_string) = @_;

    open(PRIMER, ">primer.temp") or die "Couldn't open primer.temp: $!\n";
	
    print PRIMER "PRIMER_SEQUENCE_ID=$unique_gene_id\n";
    print PRIMER "SEQUENCE=$exon_string\n";
#    print PRIMER "INCLUDED_REGION=1,400\n";
    print PRIMER "PRIMER_MIN_SIZE=18\n";
    print PRIMER "PRIMER_MAX_SIZE=22\n";
    print PRIMER "PRIMER_OPT_TM=60.0\n";
    print PRIMER "PRIMER_MIN_TM=58.0\n";
    print PRIMER "PRIMER_MAX_TM=62.0\n";
    print PRIMER "PRIMER_MIN_GC=40\n";
    print PRIMER "PRIMER_MAX_GC=60\n";
    print PRIMER "PRIMER_SALT_CONC=50.0\n";
    print PRIMER "PRIMER_DNA_CONC=50.0\n";
    print PRIMER "PRIMER_NUM_NS_ACCEPTED=0\n";
    print PRIMER "PRIMER_SELF_ANY=8\n";
    print PRIMER "PRIMER_SELF_END=3\n";
    print PRIMER "PRIMER_GC_CLAMP=0\n";
#    print PRIMER "PRIMER_PRODUCT_SIZE_RANGE=300-500\n";   # defult product size, exon > 400 
    print PRIMER "PRIMER_PRODUCT_SIZE_RANGE=250-400\n";    # set when exon > 300 used
    print PRIMER "=\n";

    close(PRIMER) or die "Couldn't close primer.temp\n";
}


=head2 primer_results

 Title   : primer_results
 Usage   : primer_results()
 Function: parses the results.temp files written
 Function: by primer3
 Example : primer_results()
 Returns : a list of array references
 Args    : none

=cut

sub primer_results
{

    my @prim_out_results = ();


    open(RESULTS, "results.temp") or die "Couldn't open results.temp: $!\n";

    while (<RESULTS>) 
    {

  	if (m/^PRIMER_ERROR=(.+)/) 
	{
  	    push(@prim_out_results, [$1]);
 	    close(RESULTS); 
	    return @prim_out_results;
	}
    }

    close(RESULTS);

      my ($id, $qual, $left, $right, $l_coord, $l_len, 
	  $r_coord, $r_len, $l_tm, $r_tm, $l_gc, $r_gc, $prod);


      open(RESULTS, "results.temp") or die "Couldn't open results.temp: $!\n";

      while (<RESULTS>) 
      {

  	chomp $_;

  	if (m/^PRIMER_SEQUENCE_ID=(\S+)/) 
  	{
  	    $id = $1;
	    print "$id\n";
  	}

	if (m/PRIMER_LEFT_SEQUENCE/) {
	  print "$_ \n";
	}

	if (m/^PRIMER_PAIR_(?:QUALIT(?:Y|Y_\d)|PENALT(?:Y|Y_\d))=(\S+)/) 
	{
	    $qual = $1;
		print "$qual\n";
	}
	if (m/^PRIMER_LEFT(?:_|_\d_)SEQUENCE=(\w+)/) 
	{
	    $left = $1;
		print "$left\n";
	}
	if (m/^PRIMER_RIGHT(?:_|_\d_)SEQUENCE=(\w+)/) 
	{
	    $right = $1;
		print "$right\n";
	}
	if (m/^PRIMER_LEF(?:T|T_\d)=(\d+),(\d+)/ )
	{
	    ($l_coord, $l_len) = ($1, $2);
#		print "$l_coord\n";
#		print "$l_len\n";
	}
	if (m/^PRIMER_RIGH(?:T|T_\d)=(\d+),(\d+)/ )
	{
	    ($r_coord, $r_len) = ($1, $2);
#		print "$r_coord\n";
#		print "$r_len\n";
	}
	if (m/^PRIMER_LEFT(?:_|_\d_)TM=(\S+)/) 
	{
	    $l_tm = $1;
#		print "$l_tm\n";
	}
	if (m/^PRIMER_RIGHT(?:_|_\d_)TM=(\S+)/) 
	{
	    $r_tm = $1;
#		print "$r_tm\n";
	}
	if (m/^PRIMER_LEF(?:T_|T_\d_)GC_PERCENT=(\d+)/)
	{
	    $l_gc = $1;
#	    print "$l_gc\n";
	}
	if (m/^PRIMER_RIGH(?:T_|T_\d_)GC_PERCENT=(\d+)/)
	{
	    $r_gc = $1;
#	    print "$r_gc\n";
	}

	if (m/^PRIMER_PRODUCT_SIZ(?:E|E_\d)=(\d+)/) 
	{
	    $prod = $1;
#		print "$prod\n\n";
	    push @prim_out_results, [$id, $qual, $left, $right, $l_coord, $l_len, 
				     $r_coord, $r_len, $l_tm, $r_tm, $l_gc, $r_gc, $prod];

	}

    }

#    for my $i (0 .. $#prim_out_results) 
#    {
#	print @{$prim_out_results[$i]}, "\n";
#    }

    close(RESULTS);

    return @prim_out_results;

}


=head2 prim_filter

 Title   : prim_filter
 Usage   : prim_filter()
 Function: sets up a filter for use with the 
 Function: Blast parsing Bioperl module
 Example : prim_filter()
 Returns : a blast score cutoff value
 Args    : none

=cut

sub prim_filter
{

    my $primhit = shift;
    return $primhit->score >= 70;

}


=head2 prim_blast_parse

 Title   : prim_blast_parse
 Usage   : prim_blast_parse(gene id, primer length, parent)
 Function: parses Blast hits against primer sequence
 Example : prim_blast_parse($exon_id, $prim_len, $parent)
 Returns : True if significant similarity is found
 Args    : gene id string, primer length string, 
 Args    : parent molecule string

=cut

sub prim_blast_parse
{

    my ($exon_id, $prim_len, $parent) = @_;
    my $blastprimObj = Bio::Tools::Blast->new(
					  -file    => 'primer_blast.out',
					  -parse   => 1,
					  -filt_func  => \&prim_filter 
					  );

    my $prim_blast_fail;

# get blast object and assign results to temp array
# then throw away the first value which is always
# a self hit

    my @hits = ($blastprimObj->hits);
    shift @hits;
    
    foreach my $hit ( @hits )
    {
	my $primalign_len = $hit->length_aln();
        my $primfrac_ident = $hit->frac_identical();
	my $prim_idscore = $hit->matches('id');

	# check for a duplicate & report failures
	if (($primalign_len == $prim_len) && ($primfrac_ident == 1.0))
	{
	    print FAILED_FILE "GENE:$exon_id\tFAILED\t PRIMER_BLAST: duplicate sequence\t (in parent:$parent)\n";
	    print LOG_FILE "GENE:$exon_id Duplicate primer sequence detected for $exon_id. Process manually\n";
	    $prim_blast_fail++;
	} 

	# significant hit? (> 90% identity over > 90% length)
	elsif (($prim_idscore / $prim_len >= 0.9) && ($primfrac_ident >= 0.9)) 
	{
	    print LOG_FILE "GENE:$exon_id\tFAILED\tPRIMER_BLAST: $prim_idscore identical in $prim_len bases\n";
#	    print "$exon_id primers failed BLAST: $prim_idscore identical in $prim_len bases\n";
	    $prim_blast_fail++;
	} 
    }

    if ($prim_blast_fail)
    {
	return 1;            
    }
    else
    {
	return 0;
    }

}


=head2 primer2parent_coord

 Title   : primer2parent_coord
 Usage   : primer2parent_coord(strand, exon start coordinate in parent, exon 
 Usage   : end coordinate in parent, forward primer coordinate, reverse primer 
 Usage   : coordinate, PCR product length)
 Function: calculates the coordinates of the primer sequences
 Function: in the parent molecule using output cordinates from primer3
 Example : primer2parent_coord($strand, $start, $end, $for_coord, $rev_coord, $prod_length)
 Returns : new primer coordinates within the parent DNA sequence
 Args    : see Usage
 
=cut

sub primer2parent_coord
{

    my ($strand, $start, $end, $for_coord, $rev_coord, $prod_length) = @_;
    my ($for_coord_inparent, $rev_coord_inparent);

  if ($strand == 1) 
  {
	$for_coord_inparent = $start + $for_coord -1;
	$rev_coord_inparent = $for_coord_inparent + $prod_length -1;
#	print "$strand: for:$for_coord_inparent = $start + $for_coord, \n";
#	print "rev:$rev_coord_inparent = $for_coord_inparent + $prod_length\n\n";
    } 
    else 
    {
	$rev_coord_inparent = $end - $for_coord -1;
	$for_coord_inparent = $rev_coord_inparent - $prod_length +1;
#	print "$strand: for:$for_coord_inparent = $rev_coord_inparent - $prod_length\n";
#	print "rev:$rev_coord_inparent = $end - $for_coord\n\n";
    }

  return ($for_coord_inparent, $rev_coord_inparent);

}


=head2 extract_pcr_products

 Title   : extract_pcr_products
 Usage   : extract_pcr_products(list of exon data, 
 Usage   : primer coordinates in parent)
 Function: extracts the PCR product from the parent DNA sequence
 Example : extract_pcr_products($exon_data, $pcr_prod_start, $pcr_prod_end)
 Returns : string of parent DNA sequence
 Args    : list of exon data, primer coordinates in parent
 
=cut

sub extract_pcr_products
{

    my ($exon_data, $pcr_prod_start, $prod_length, $strand) = @_;
    #my $pcr_product_string = $exon_data->subseq($pcr_prod_start, $pcr_prod_end);
   my $pcr_product_string .= substr($exon_data->{str}{val}, $pcr_prod_start, $prod_length);
    if ($strand == -1) {
      $pcr_product_string = reverse($pcr_product_string);
      $pcr_product_string =~ tr/[GTACgtac]/[CATGcatg]/;
}
    return $pcr_product_string;

}


=head2 extract_pcr_products

 Title   : extract_pcr_products
 Usage   : extract_pcr_products(list of exon data, 
 Usage   : primer coordinates in parent)
 Function: extracts the PCR product from the parent DNA sequence
 Example : extract_pcr_products($exon_data, $pcr_prod_start, $pcr_prod_end)
 Returns : string of parent DNA sequence
 Args    : list of exon data, primer coordinates in parent
 
=cut
    
sub write_tmp_fasta
{

    my ($seq_file, $gene_id) = @_;

    my $temp_product = Bio::Seq->new
	(
	 -seq            => $seq_file,
	 -display_id     => $gene_id
	 );

    my $db_out = Bio::SeqIO->new(-file => ">temp.dna", '-format' => 'Fasta');

    $db_out->write_seq($temp_product);

}

sub filter
{

    my $hit = shift;
    return $hit->score >= 400; # original cutoff
#    return $hit->score >= 1000;  # relaxed cutoff

}

sub blast_parse
{
    system("touch BLAST_failures");
    my ($exon_id, $pcrprod_len, $prim_pair_count, $num_prim_pairs, $parent) = @_;
    my $blastObj = Bio::Tools::Blast->new(
					  -file    => 'temp_blast.out',
					  -parse   => 1,
					  -filt_func  => \&filter 
					  );

    my $blast_fail;
    my $duplicate = 0;

# get blast object and assign results to temp array
# then throw away the first value which is always
# a self hit

    my @hits = ($blastObj->hits);
    shift @hits;
    
#   foreach my $hit ($blastObj->hits)

    foreach my $hit ( @hits )
    {
	my $hit_name = $hit->name();
        my $score = $hit->score();
	my $align_len = $hit->length_aln();
	my $frac_ident = $hit->frac_identical();

	print "$exon_id ------------------------->\n";
	print "Hit:$hit_name, Ali:$align_len, Prod:$pcrprod_len, Ide:$frac_ident, Sco:$score\n\n";

	if ($align_len == $pcrprod_len and $frac_ident == 1.0) {
	    $duplicate++;
	} 
	elsif ($score) 
	{
	    print LOG_FILE "GENE:$exon_id NON-UNIQUE BLAST:$hit_name has a significant score value:$score\n";
	    system("cat temp_blast.out >> BLAST_failures");
	    $blast_fail++;
	    last;
	} 
    }

    if ($duplicate > 1) 
    {
	print FAILED_FILE "GENE:$exon_id\tFAILED\t AMPLICON_BLAST: duplicate sequence\t (in parent:$parent)\n";
	print LOG_FILE "GENE:$exon_id WARNING:Duplicate sequence detected. Process manually\n";
	$blast_fail++;
    }

    if (($blast_fail) && ($prim_pair_count == $num_prim_pairs)) 
    {
#	print FAILED_FILE "GENE:$exon_id\tFAILED\t BLAST_AMPLICON: All primers failed\t (in parent:$parent)\n";
	print LOG_FILE "GENE:$exon_id  NON-UNIQUE BLAST: All primer pairs failed.\n";
    }

    $blastObj->destroy();

    if ($blast_fail)
    {
	return 0;
    }
    else
    {
	return 1;
    }


}
