#!/usr/local/bin/perl5.6.0 -w

=head1 NAME

sp2embl.pl

=head1 SYNOPSIS

  sp2embl.pl 

  -help a help message
  -fa the proteins input file
  -sp the superfamily file
  -embl file, including the DNA sequence and 0 or more features

=head1 DESCRIPTION

  This script generates a new embl file with SUPERFAMILY predicted Features

  @author axk@sanger.ac.uk 

=cut

# External Modules

# Constraint variables declaration before using them
use strict;
use Cwd;
use Data::Dumper;

# Benchmark Module
use Benchmark;
##################################################################

# Bio modules
# PSU
use Bio::PSU::SeqFactory;

##########
# The Superfamily Parser Module
##

eval "require Bio::PSU::Utils::SuperfamilyParser";
# use Bio::PSU::Utils::SuperfamilyParser;

##########
# The Protein to DNA Coordinates Mapping Module
##

eval "require Bio::PSU::Utils::MapProteinDNA_coordinates";
# use Bio::PSU::Utils::MapProteinDNA_coordinates;

my $t1 = Benchmark->new ();

##########################################
#
# Verbose
#
##########################################

my $verbose = 0;

##########################################
#
# Loading Parameters
#
##########################################

my $_help        = "-help";
my $_superfamily = "-sp";
my $_fasta       = "-fa";
my $_embl        = "-embl";

my $sp   = "-1";
my $fa   = "-1";
my $embl = "-1";

# Check if in the command line there is an input file name (-i option)
# the other options are not compulsory (set of options per default)
my $lgth = @ARGV + 0;
# if no args, display the help message
if ($lgth < 6) {
  display_help ();
  exit 1;
}
elsif ($lgth >= 6) {
  my $i = 0;
  foreach my $element (@ARGV) {
    if ($element eq $_help) {
      display_help ();
      exit 1;
    }
    elsif ($element eq $_superfamily) {
      $sp = $ARGV[$i+1];
    }
    elsif ($element eq $_fasta) {
      $fa = $ARGV[$i+1];
    }
    elsif ($element eq $_embl) {
      $embl = $ARGV[$i+1];
    }
    # if at least one of the options is not recognized, display the help message and exit
    elsif ($element =~ /^-/) {
      print STDERR "option not recognized: $element\n";
      display_help ();
      exit 1;
    }
    $i ++;
  }
}
if (($sp eq "-1") || ($fa eq "-1") || ($embl eq "-1")) {
  display_help ();
  exit 1;
}

####################
# Generate an Array of embl files
# ($embl may be a directory)
##

my @embls = ();

if (-f $embl) {
  push (@embls, $embl);
}
else {
  opendir EMBLDIR, $embl or die "Can not read this directory, $embl!\n";
  @embls = grep {$_ =~ /.+\.embl$/} readdir EMBLDIR;

  closedir EMBLDIR;
}

if ($verbose) {
  print STDERR "set of embl files: ". join (', ', @embls) . "\n";
}

####################
# parsing proteins fasta file
##

if ($verbose) {
  print STDERR "parsing fasta sequence file, $fa...\n";
}

my %prots;
my $sfa = Bio::PSU::SeqFactory->make(-file   => "<$fa",
				     -format => 'fasta',
				     -type   => 'protein',
				    );
############
# check if $prot is unique !!!!!!!!
##

while (my $seq = $sfa->next_seq) {
  my $prot   = $seq->id;
  my $range  = $seq->desc;
  my $length = length ($seq->str);
  $prot =~ /(DICTY\d+P\d+_\d+).+/;
  my $contig = $1;

  if ($verbose) {
    print STDERR "parsing protein fasta sequence, $prot...\n";
    # print STDERR "found the following description, $range.\n";
    # print STDERR "found the following contig name, $contig.\n";
  }

  my %prot = (
	      contig => $contig,
	      range  => $range,
	      length => $length,
	     );
  $prots{$prot}=\%prot;
}

if ($verbose) {
  print STDERR "parsing done.\nChecking prots hash: " . Dumper (%prots) . "\n";
}

#########################
# get SP Predictions
##

if ($verbose) {
  print STDERR "parsing SP assignments, $sp...\n";
}

my %sp_fts_per_contig;

my $sp_parser = Bio::PSU::Utils::SuperfamilyParser->new (
							 superfamilyOutput => $sp,
							 verbose => $verbose,
							);
my @hits = $sp_parser->parse ();

if ($verbose) {
  # print STDERR "found " . @hits . " hits. " . Dumper (@hits) . "\n";;
}

foreach my $sp_hit (@hits) {
  my $prot      = $sp_hit->getProtAccession();
  my $start     = $sp_hit->getStart();
  my $end       = $sp_hit->getEnd();
  my $evalue    = $sp_hit->getEvalue();
  my $scop_id   = $sp_hit->getScopID();
  my $scop_name = $sp_hit->getScopName();
 
  $prot =~ /(DICTY\d+P\d+_\d+).+/;
  my $contig = $1;
  my %prot = %{$prots{$prot}};
  my $ranges_str = $prot{range};
  my $protein_length = $prot{length};
  my @ranges = parse_ranges_string ($ranges_str);

  # Mapping Protein coordinates to DNA coordinates of the SP domain

  my $mapping = Bio::PSU::Utils::MapProteinDNA_coordinates->new (
								 sp_start => $start,
								 sp_end  => $end,
								 protein_length => $protein_length,
								 gene_ranges => \@ranges,
								);
  my @sp_ranges = $mapping->mapRange ();

  if ($verbose) {
    print STDERR "mapped SP domain coordinates ($end - $start) into the genomic sequence: " . Dumper (@sp_ranges) . "\n";
    print STDERR "given protein coordinates: " . Dumper (@ranges) . "\n";
  }

  my $feature_name = "CDS_domain";
  my $note = "SUPERFAMILY domain match for $prot\t(P=$evalue)\tto SCOP entry, $scop_id $scop_name from $start to $end";
  my $colour = 7;

  my $sp_ft = create_sp_ft ($feature_name, $note, $colour, @sp_ranges);
  if (exists $sp_fts_per_contig{$contig}) {
    my @sp_fts = @{$sp_fts_per_contig{$contig}};
    push (@sp_fts, $sp_ft);
    $sp_fts_per_contig{$contig} = \@sp_fts;
  }
  else {
    my @sp_fts = ();
    push (@sp_fts, $sp_ft);
    $sp_fts_per_contig{$contig} = \@sp_fts;
  }
}

if ($verbose) {
  print STDERR "Parsing done.\nChecking sp_fts_per_contig hash: " . Dumper (%sp_fts_per_contig) . "\n";
}

##########################
# parsing embl file(s)
#
# NB: if no SP assignments, no need to parse embl file => just copy it !!
##

foreach my $embl_file (@embls) {
  
  if ($verbose) {
    print STDERR "parsing embl file, $embl_file...\n";
  }
  
  $embl_file =~ /(.+)\.embl/;
  my $embl_contig = $1;
  if ($verbose) {
    print STDERR "contig from embl file parsing: $embl_contig\n";
  }

  my $sembl = Bio::PSU::SeqFactory->make(-file   => "<$embl_file",
					 -format => 'embl'
					);
  
  my $embl_seq_obj = $sembl->next_seq;
  my @features = ();
  if ((my $fnum = $embl_seq_obj->has_features) > 0) {
    foreach my $feature ($embl_seq_obj->features) {
      push (@features, $feature);
    }
  }
  
  if ($verbose) {
    print STDERR "parsing done.\n";
  }
  
  # add sp features

  if (exists $sp_fts_per_contig{$embl_contig}) {
    push (@features, @{$sp_fts_per_contig{$embl_contig}});
  }
  else {
    if ($verbose) {
      print STDERR "no superfamily prediction for contig, $embl_contig\n";
    }
  }

  if ($verbose) {
    print STDERR "features array: " . Dumper (@features) . "\n";
  }
  
  ##################
  # generate new embl file
  ##
  
  if ($verbose) {
    print STDERR "generating new embl file with the SP assignment features...\n";
  }
  
  my $new_embl_seq_obj = Bio::PSU::Seq->new (-id   => $embl_seq_obj->id,
					     -desc => $embl_seq_obj->desc,
					     -type => $embl_seq_obj->type,
					     -str  => $embl_seq_obj->str
					    );
  $new_embl_seq_obj->features (@features);
  my $sembl_new = Bio::PSU::SeqFactory->make(-file   => ">$embl_file.new",
					     -format => 'embl'
					    );
  $sembl_new->write_seq($new_embl_seq_obj);
  
}

print STDERR "process done.\n";

my $t2 = Benchmark->new ();
print  STDERR "\nTotal : ", timestr (timediff ($t2, $t1)), "\n";

###################################################################
# The end
###################################################################

# display the help message

sub display_help {
  print "sp2embl.pl -help -fa [protein fasta file] -sp [superfamily file] -embl [DNA sequence embl file]\n\n";
  print "\t-help a help message\n";
  print "\t-fa the protein input file - (no default, always specify it)\n";
  print "\t-sp the superfamily prediction file - (no default, always specify it)\n";
  print "\t-embl the DNA sequence input file - (no default, always specify it)\n\n";
}

sub create_sp_ft {
  my ($feature_name, $note, $colour, @sp_ranges) = @_;
  
  my $feature = Bio::PSU::Feature->new(
				       -key    => $feature_name,
				       -ranges => \@sp_ranges,
				      );
  $feature->qadd('note', $note);
  $feature->qadd('colour', $colour);
  return $feature;
}

sub parse_ranges_string {
  my ($ranges_str) = @_;
  my $strand = 1;
  my @ranges = ();
  
  if ($ranges_str =~ /complement\((.+)\)/) {
    $strand = -1;
    $ranges_str = $1;
  }
  
  while ($ranges_str =~ /(\d+)\.\.(\d+),(.+)/) {
    my $start = $1;
    my $end   = $2;
    $ranges_str = $3;
    
    my $range = Bio::PSU::Range->new(-start  => $start,
				     -end    => $end,
				     -strand => $strand,
				    );
    push (@ranges, $range);
  }
  
  $ranges_str =~ /(\d+)\.\.(\d+)/;
  my $start = $1;
  my $end   = $2;
  
  my $range = Bio::PSU::Range->new(-start  => $start,
				   -end    => $end,
				   -strand => $strand,
				  );
  push (@ranges, $range);
  
  if ($verbose) {
    print STDERR "ranges_str, $ranges_str, parsing res: " . Dumper (@ranges) . "\n";
  }
  
  return @ranges;
}
